import torch.nn as nn
import torch.nn.functional as F

from antgo.framework.helper.models.builder import LOSSES
from .utils import weighted_loss


@weighted_loss
def knowledge_distillation_kl_div_loss(pred,
                                       soft_label,
                                       T,
                                       detach_target=True):
    r"""Loss function for knowledge distilling using KL divergence.

    Args:
        pred (Tensor): Predicted logits with shape (N, n + 1).
        soft_label (Tensor): Target logits with shape (N, N + 1).
        T (int): Temperature for distillation.
        detach_target (bool): Remove soft_label from automatic differentiation

    Returns:
        torch.Tensor: Loss tensor with shape (N,).
    """
    assert pred.size() == soft_label.size()
    target = F.softmax(soft_label / T, dim=1)
    if detach_target:
        target = target.detach()

    kd_loss = F.kl_div(
        F.log_softmax(pred / T, dim=1), target, reduction='none').mean(1) * (
            T * T)

    return kd_loss


@LOSSES.register_module()
class KnowledgeDistillationKLDivLoss(nn.Module):
    """Loss function for knowledge distilling using KL divergence.

    Args:
        reduction (str): Options are `'none'`, `'mean'` and `'sum'`.
        loss_weight (float): Loss weight of current loss.
        T (int): Temperature for distillation.
    """

    def __init__(self, reduction='mean', loss_weight=1.0, T=10):
        super(KnowledgeDistillationKLDivLoss, self).__init__()
        assert T >= 1
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.T = T

    def forward(self,
                pred,
                soft_label,
                weight=None,
                avg_factor=None,
                reduction_override=None):
        """Forward function.

        Args:
            pred (Tensor): Predicted logits with shape (N, n + 1).
            soft_label (Tensor): Target logits with shape (N, N + 1).
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Defaults to None.
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')

        reduction = (
            reduction_override if reduction_override else self.reduction)

        loss_kd = self.loss_weight * knowledge_distillation_kl_div_loss(
            pred,
            soft_label,
            weight,
            reduction=reduction,
            avg_factor=avg_factor,
            T=self.T)

        return loss_kd



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from functools import reduce
import paddle
from paddle.fluid import framework, unique_name
from paddle.fluid.core import VarDesc
from paddle.fluid.initializer import XavierInitializer, MSRAInitializer
from paddle.fluid.data_feeder import check_variable_and_dtype
from antgo.framework.torch2paddle.utils import paddle_dtypes


def _calculate_fan_in_and_fan_out(var):
    dimensions = var.dim()
    if dimensions < 2:
        raise ValueError(
            "Fan in and fan out can not be computed for var with fewer than 2 dimensions"
        )
    num_input_fmaps = var.shape[0]
    num_output_fmaps = var.shape[1]
    receptive_field_size = 1
    if var.dim() > 2:
        receptive_field_size = reduce(lambda x, y: x * y, var.shape[2:])
    fan_in = num_input_fmaps * receptive_field_size
    fan_out = num_output_fmaps * receptive_field_size
    return fan_in, fan_out


def _calculate_correct_fan(var, mode):
    mode = mode.lower()
    valid_modes = ['fan_in', 'fan_out']
    if mode not in valid_modes:
        raise ValueError("Mode {} not supported, please use one of {}".format(
            mode, valid_modes))
    fan_in, fan_out = _calculate_fan_in_and_fan_out(var)
    return fan_in if mode == 'fan_in' else fan_out


def _calculate_gain(nonlinearity, param=None):
    linear_fns = [
        'linear', 'conv1d', 'conv2d', 'conv3d', 'conv_transpose1d',
        'conv_transpose2d', 'conv_transpose3d'
    ]
    if nonlinearity in linear_fns or nonlinearity == 'sigmoid':
        return 1
    elif nonlinearity == 'tanh':
        return 5.0 / 3
    elif nonlinearity == 'relu':
        return math.sqrt(2.0)
    elif nonlinearity == 'leaky_relu':
        if param is None:
            negative_slope = 0.01
        elif not isinstance(param, bool) and isinstance(
                param, int) or isinstance(param, float):
            # True/False are instances of int, hence check above
            negative_slope = param
        else:
            raise ValueError("negative_slope {} not a valid number".format(
                param))
        return math.sqrt(2.0 / (1 + negative_slope**2))
    elif nonlinearity == 'selu':
        return 3.0 / 4  # Value found empirically (https://github.com/pytorch/pytorch/pull/50664)
    else:
        raise ValueError("Unsupported nonlinearity {}".format(nonlinearity))


class KaimingNormal(MSRAInitializer):
    def __init__(self, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        super(KaimingNormal, self).__init__(uniform=False, fan_in=None, seed=0)
        self.a = a
        self.mode = mode
        self.nonlinearity = nonlinearity

    def __call__(self, var, block=None):
        """Initialize the input tensor with MSRA initialization.
        Args:
            var(Tensor): Tensor that needs to be initialized.
            block(Block, optional): The block in which initialization ops
                   should be added. Used in static graph only, default None.
        Returns:
            The initialization op
        """
        block = self._check_block(block)

        assert isinstance(var, framework.Variable)
        assert isinstance(block, framework.Block)
        f_in, f_out = self._compute_fans(var)

        if self._seed == 0:
            self._seed = block.program.random_seed

        # to be compatible of fp16 initalizers
        if var.dtype == paddle_dtypes.t_float16:
            out_dtype = paddle_dtypes.t_float32
            out_var = block.create_var(
                name=unique_name.generate(".".join(
                    ['masra_init', var.name, 'tmp'])),
                shape=var.shape,
                dtype=out_dtype,
                type=VarDesc.VarType.LOD_TENSOR,
                persistable=False)
        else:
            out_dtype = var.dtype
            out_var = var

        fan = _calculate_correct_fan(var, self.mode)
        gain = _calculate_gain(self.nonlinearity, self.a)
        std = gain / math.sqrt(fan)
        op = block._prepend_op(
            type="gaussian_random",
            outputs={"Out": out_var},
            attrs={
                "shape": out_var.shape,
                "dtype": int(out_dtype),
                "mean": 0.0,
                "std": std,
                "seed": self._seed
            },
            stop_gradient=True)

        if var.dtype == VarDesc.VarType.FP16:
            block.append_op(
                type="cast",
                inputs={"X": out_var},
                outputs={"Out": var},
                attrs={"in_dtype": out_var.dtype,
                       "out_dtype": var.dtype})

        if not framework.in_dygraph_mode():
            var.op = op
        return op


def kaiming_normal_(param, a=0, mode='fan_in', nonlinearity='leaky_relu'):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=KaimingNormal(
            a=a, mode=mode, nonlinearity=nonlinearity))
    paddle.assign(replaced_param, param)


class XavierNormal(XavierInitializer):
    def __init__(self, gain=1.0):
        super(XavierNormal, self).__init__(
            uniform=True, fan_in=None, fan_out=None, seed=0)
        self._gain = gain

    def __call__(self, var, block=None):
        block = self._check_block(block)
        assert isinstance(block, framework.Block)
        check_variable_and_dtype(var, "Out", ["float16", "float32", "float64"],
                                 "xavier_init")

        fan_in, fan_out = _calculate_fan_in_and_fan_out(var)

        if self._seed == 0:
            self._seed = block.program.random_seed

        # to be compatible of fp16 initalizers
        if var.dtype == paddle_dtypes.t_float16:
            out_dtype = paddle_dtypes.t_float32
            out_var = block.create_var(
                name=unique_name.generate(".".join(
                    ['xavier_init', var.name, 'tmp'])),
                shape=var.shape,
                dtype=out_dtype,
                type=VarDesc.VarType.LOD_TENSOR,
                persistable=False)
        else:
            out_dtype = var.dtype
            out_var = var

        std = self._gain * math.sqrt(2.0 / float(fan_in + fan_out))
        op = block._prepend_op(
            type="uniform_random",
            inputs={},
            outputs={"Out": out_var},
            attrs={
                "shape": out_var.shape,
                "dtype": out_dtype,
                "min": 0,
                "max": std,
                "seed": self._seed
            },
            stop_gradient=True)
        if var.dtype == paddle_dtypes.t_float16:
            block.append_op(
                type="cast",
                inputs={"X": out_var},
                outputs={"Out": var},
                attrs={"in_dtype": out_var.dtype,
                       "out_dtype": var.dtype})
        if not framework.in_dygraph_mode():
            var.op = op
        return op


def xavier_normal_(param, gain=1.0):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=XavierNormal(gain=gain))
    paddle.assign(replaced_param, param)


class XavierUniform(XavierInitializer):
    def __init__(self, gain=1.0):
        super(XavierUniform, self).__init__(
            uniform=True, fan_in=None, fan_out=None, seed=0)
        self._gain = gain

    def __call__(self, var, block=None):
        block = self._check_block(block)
        assert isinstance(block, framework.Block)
        check_variable_and_dtype(var, "Out", ["float16", "float32", "float64"],
                                 "xavier_init")

        fan_in, fan_out = _calculate_fan_in_and_fan_out(var)

        if self._seed == 0:
            self._seed = block.program.random_seed

        # to be compatible of fp16 initalizers
        if var.dtype == paddle_dtypes.t_float16:
            out_dtype = paddle_dtypes.t_float32
            out_var = block.create_var(
                name=unique_name.generate(".".join(
                    ['xavier_init', var.name, 'tmp'])),
                shape=var.shape,
                dtype=out_dtype,
                type=VarDesc.VarType.LOD_TENSOR,
                persistable=False)
        else:
            out_dtype = var.dtype
            out_var = var

        std = self._gain * math.sqrt(2.0 / float(fan_in + fan_out))
        limit = math.sqrt(3.0) * std
        op = block._prepend_op(
            type="uniform_random",
            inputs={},
            outputs={"Out": out_var},
            attrs={
                "shape": out_var.shape,
                "dtype": out_dtype,
                "min": -limit,
                "max": limit,
                "seed": self._seed
            },
            stop_gradient=True)
        if var.dtype == paddle_dtypes.t_float16:
            block.append_op(
                type="cast",
                inputs={"X": out_var},
                outputs={"Out": var},
                attrs={"in_dtype": out_var.dtype,
                       "out_dtype": var.dtype})
        if not framework.in_dygraph_mode():
            var.op = op
        return op


def xavier_uniform_(param, gain=1.0):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=XavierUniform(gain=gain))
    paddle.assign(replaced_param, param)


def constant_init_(param, val):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.full(param.shape, val, param.dtype)))
    paddle.assign(replaced_param, param)


def normal_init_(param, mean=0.0, std=1.0):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.normal(
                mean=mean, std=std, shape=param.shape)))
    paddle.assign(replaced_param, param)


def ones_init_(param):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.ones(param.shape, param.dtype)))
    paddle.assign(replaced_param, param)


def zeros_init_(param):
    replaced_param = paddle.create_parameter(
        shape=param.shape,
        dtype=param.dtype,
        default_initializer=paddle.nn.initializer.Assign(
            paddle.zeros(param.shape, param.dtype)))
    paddle.assign(replaced_param, param)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/18 12:48
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from .runas_op import runas_op
from .inference_model_op import inference_model_op
from .inference_onnx_op import inference_onnx_op, ensemble_onnx_op, ensemble_shell_op

__all__ = [
    'runas_op', 'inference_model_op', 'inference_onnx_op', 'ensemble_onnx_op', 'ensemble_shell_op'
]



# encoding=utf-8
# @Time    : 17-6-13
# @File    : recorder.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.core import *
import copy
import numpy as np
import os
import json
try:
  import queue
except:
  import Queue as queue
import multiprocessing
from antgo.task.task import *
from antgo.dataflow.basic import *
import requests
from antgo.context import *
from antgo.utils import *
import cv2


class RecorderNode(Node):
  def __init__(self, inputs):
    super(RecorderNode, self).__init__(name=None, action=self.action, inputs=inputs, auto_trigger=True)
    self._dump_dir = None
    self._annotation_cache = queue.Queue()
    self._record_writer = None
    self._is_none = False
  
  @property
  def is_measure(self):
    if self._record_writer is None:
      return False
    
    if self._record_writer.size == 0:
      return False
    
    if self._is_none:
      return False
    
    return True

  def close(self):
    if self._record_writer is not None:
        self._record_writer.close()
    self._record_writer = None
    
  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val
    
    if self._dump_dir is None:
      self.close()
      return
      
    # remove existed dump_dir
    if os.path.exists(self._dump_dir):
      shutil.rmtree(self._dump_dir)
    
    # mkdir
    os.makedirs(self._dump_dir)
    
    # set record workspace
    self._record_writer = RecordWriter(self._dump_dir)

  def action(self, *args, **kwargs):
    value = [copy.deepcopy(args[0])]
    for entry in value:
      self._annotation_cache.put(copy.deepcopy(entry))

  def record(self, val, **kwargs):
    results = []
    results_label = []
    if type(val) == list or type(val) == tuple:
      for aa in val:
        if type(aa) == dict:
          results.append(aa['RESULT'])
          aa.pop('RESULT')
          results_label.append(aa)
        else:
          results.append(aa)
    else:
      if type(val) == dict:
        results.append(val['RESULT'])
        val.pop('RESULT')
        results_label.append(val)
      else:
        results = [val]
    
    for index, result in enumerate(results):
      gt = None
      if self._annotation_cache.qsize() > 0:
        gt = self._annotation_cache.get()
      
      if gt is None and not self._is_none:
        self._is_none = True
      
      if len(results_label) > 0:
        self._record_writer.write(Sample(groundtruth=gt, predict=result, predict_label=results_label[index]))
      else:
        self._record_writer.write(Sample(groundtruth=gt, predict=result, predict_label=[]))

  def iterator_value(self):
    pass


class RecorderNode2(object):
  def __init__(self):
    self._dump_dir = None
    self._record_writer = None
  
  @property
  def is_measure(self):
    if self._record_writer is None:
      return False
    
    if self._record_writer.size == 0:
      return False    
    return True

  def close(self):
    if self._record_writer is not None:
        self._record_writer.close()
    self._record_writer = None
    
  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val
    
    if self._dump_dir is None:
      self.close()
      return
      
    # remove existed dump_dir
    if os.path.exists(self._dump_dir):
      shutil.rmtree(self._dump_dir)
    
    # mkdir
    os.makedirs(self._dump_dir)
    
    # set record workspace
    self._record_writer = RecordWriter(self._dump_dir)

  def record(self, val, **kwargs):
    # KEY: PREDICT, KEY: GT
    predicts = []
    gts = []
    if type(val) == list or type(val) == tuple:
      for a in val:
        assert('PREDICT' in a and 'GT' in a)
        predicts.append(a['PREDICT']['data'])
        gts.append(a['GT']['data'])
    else:
      assert('PREDICT' in val and 'GT' in val)
      predicts = [val['PREDICT']['data']]
      gts = [val['GT']['data']]

    for predict, gt in zip(predicts, gts):
      self._record_writer.write(Sample(groundtruth=gt, predict=predict, predict_label=[]))


class QueueRecorderNode(Node):
  def __init__(self, inputs, output_queue):
    super(QueueRecorderNode, self).__init__(name=None, action=self.action, inputs=inputs,auto_trigger=True)

    self._annotation_cache = queue.Queue()

    self.recorder_output_queue = output_queue
    self._dump_dir = None
    self._is_none = False
    
    setattr(self,'model_fn', None)

  def _transfer_data(self, result, key):
      transfer_result = None
      transfer_result_type = None

      if result['%s_TYPE'%key] in ['SCALAR','FILE', 'STRING']:
        transfer_result = str(result[key])
        transfer_result_type = result['%s_TYPE'%key]
      elif result['%s_TYPE'%key] == 'JSON':
        transfer_result = json.dumps(result[key])
        transfer_result_type = 'STRING'
      elif result['%s_TYPE'%key] == 'IMAGE':
        data = result[key]
        assert(len(data.shape) <= 3 and len(data.shape) >= 2)

        if len(data.shape) == 2:
          if data.dtype == np.uint8:
            transfer_result = data
          else:
            data_min = np.min(data)
            data_max = np.max(data)
            transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)
        else:
          assert(data.shape[2] == 3)
          transfer_result = data.astype(np.uint8)

        # save path
        if not os.path.exists(self.dump_dir):
          os.makedirs(self.dump_dir)

        image_path = os.path.join(self.dump_dir, '%s.png'%str(uuid.uuid4()))
        # scipy.misc.imsave(image_path, transfer_result)
        # imageio.imwrite(image_path, transfer_result)
        cv2.imwrite(image_path, transfer_result)
        transfer_result = image_path
        transfer_result_type = 'IMAGE'
      elif result['%s_TYPE'%key] == 'VIDEO':
        data = result[key]
        assert(type(data) == list)

        # save path
        video_path = os.path.join(self.dump_dir, '%s.mp4'%str(uuid.uuid4()))
        # writer = imageio.get_writer(video_path, fps=30)
        # for im in data:
        #   writer.append_data(im.astype(np.uint8))
        # writer.close()

        transfer_result = video_path
        transfer_result_type = 'VIDEO'
      else:
        logger.error('AUDIO not support')

      return transfer_result, transfer_result_type

  def record(self, val, **kwargs):
    results = []
    results_label = []
    if type(val) == list or type(val) == tuple:
      for aa in val:
        if 'RESULT' in aa and 'RESULT_TYPE' in aa:
          results.append({'RESULT': aa['RESULT'], 'RESULT_TYPE': aa['RESULT_TYPE']})
          aa.pop('RESULT')
          aa.pop('RESULT_TYPE')
        else:
          results.append({'RESULT': None, 'RESULT_TYPE': None})

        results_label.append(aa)
    else:
      if type(val) == dict:
        if 'RESULT' in val and 'RESULT_TYPE' in val:
          results.append({'RESULT': val['RESULT'], 'RESULT_TYPE': val['RESULT_TYPE']})
          val.pop('RESULT')
          val.pop('RESULT_TYPE')
        else:
          results.append({'RESULT': None, 'RESULT_TYPE': None})

        results_label.append(val)
      else:
        results = [{'RESULT': val, 'RESULT_TYPE': 'IMAGE'}]

    for index, result in enumerate(results):
      gt = None
      if self._annotation_cache.qsize() > 0:
        gt = self._annotation_cache.get()

      if gt is None and not self._is_none:
        self._is_none = True

      # result 中保存模型输出的唯一结果
      # results_label 中保存其他附加结果
      # 均转换成文件或字符串形式输出 [{'TYPE': 'FILE', 'PATH': ''},
      #                           {'TYPE': 'JSON', 'CONTENT': ''},
      #                           {'TYPE': 'STRING', 'CONTENT': ''},
      #                           {'TYPE': 'IMAGE', 'PATH': ''},
      #                           {'TYPE': 'VIDEO', 'PATH': ''},
      #                           {'TYPE': 'AUDIO', 'PATH': ''}]

      # 1.step for main results
      transfer_result = None
      transfer_result_type = None
      if result['RESULT'] is not None and result['RESULT_TYPE'] is not None:
        assert(result['RESULT_TYPE'] in ['FILE', 'JSON', 'SCALAR', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'])
        try:
          transfer_result, transfer_result_type = self._transfer_data(result, 'RESULT')
        except:
          transfer_result = None
          transfer_result_type = None

      # 2.step for additional results
      transfer_additional_results = []
      for k,v in results_label[index].items():
        if '_TYPE' not in k:
          if '%s_TYPE'%k in results_label[index]:
            a = None
            b = None
            try:
              a, b = self._transfer_data({k: v, '%s_TYPE'%k: results_label[index]['%s_TYPE'%k]}, k)
            except:
              a = None
              b = None

            if a is not None and b is not None:
              transfer_additional_results.append({k: a, 'TYPE': b})

      if len(results_label) > 0:
        self.recorder_output_queue.put((-1, ({'DATA': transfer_result,
                                                'TYPE': transfer_result_type},transfer_additional_results)))
      else:
        self.recorder_output_queue.put((-1, {'DATA': transfer_result,
                                               'TYPE': transfer_result_type}))

  def action(self, *args, **kwargs):
    value = copy.deepcopy(args[0])
    if type(value) != list:
      value = [value]

    for entry in value:
      self._annotation_cache.put(copy.deepcopy(entry))

  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val

  @property
  def is_measure(self):
    if self._record_writer is None:
      return False

    if self._is_none:
      return False

    return True

  def iterator_value(self):
    pass


class LocalRecorderNode(Node):
  def __init__(self, inputs):
    super(LocalRecorderNode, self).__init__(name=None, action=self.action, inputs=inputs, auto_trigger=True)
    self._annotation_cache = queue.Queue()

    self._dump_dir = None
    self._is_none = False

    self._count = 0
    setattr(self, 'model_fn', None)

  def save(self, data, data_type, name, count, index):
    # only support 'FILE', 'JSON', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'
    assert(data_type in ['FILE', 'JSON', 'SCALAR', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'])

    if data_type == 'FILE':
      if not os.path.exists(data):
        return

      norm_path = os.path.normpath(data)
      shutil.copy(data, os.path.join(self.dump_dir, '%d_%d_%s_%s'%(count, index,name,norm_path.split('/')[-1])))
    elif data_type == 'JSON':
      with open(os.path.join(self.dump_dir, '%d_%d_%s.json'%(count, index,name)),'w') as fp:
        fp.write(json.dumps(data))
    elif data_type == 'STRING' or data_type == 'SCALAR':
      with open(os.path.join(self.dump_dir, '%d_%d_%s.txt'%(count, index,name)), 'w') as fp:
        fp.write(str(data))
    elif data_type == 'IMAGE':
      transfer_result = data
      if len(data.shape) == 2:
        if data.dtype == np.uint8:
          transfer_result = data
        else:
          data_min = np.min(data)
          data_max = np.max(data)
          transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)
      else:
        assert (data.shape[2] == 3)
        transfer_result = data.astype(np.uint8)

      # scipy.misc.imsave(os.path.join(self.dump_dir, '%d_%d_%s.png'%(count, index, name)), transfer_result)
      # imageio.imwrite(os.path.join(self.dump_dir, '%d_%d_%s.png'%(count, index, name)), transfer_result)
      cv2.imwrite(os.path.join(self.dump_dir, '%d_%d_%s.png'%(count, index, name)), transfer_result)
    elif data_type == 'VIDEO':
      video_path = os.path.join(self.dump_dir, '%d_%d_%s.mp4' % (count, index, name))
      # writer = imageio.get_writer(video_path, fps=30)
      # for im in data:
      #   writer.append_data(im.astype(np.uint8))
      # writer.close()
    elif data_type == 'AUDIO':
      logger.error('now, dont support save audio')

  def action(self, *args, **kwargs):
    value = copy.deepcopy(args[0])
    if type(value) != list:
      value = [value]

    for entry in value:
      self._annotation_cache.put(copy.deepcopy(entry))

  def record(self, val, **kwargs):
    results = val
    if type(val) != list and type(val) != tuple:
      results = [val]

    for index, result in enumerate(results):
      gt = None
      if self._annotation_cache.qsize() > 0:
        gt = self._annotation_cache.get()

      if gt is None and not self._is_none:
        self._is_none = True

      # 均转换成文件或字符串形式输出 [{'TYPE': 'FILE', 'PATH': ''},
      #                           {'TYPE': 'JSON', 'CONTENT': ''},
      #                           {'TYPE': 'STRING', 'CONTENT': ''},
      #                           {'TYPE': 'IMAGE', 'PATH': ''},
      #                           {'TYPE': 'VIDEO', 'PATH': ''},
      #                           {'TYPE': 'AUDIO', 'PATH': ''}]

      for k, v in result.items():
        if not k.endswith('_TYPE'):
          if '%s_TYPE' % k in result:
            data = v
            data_type = result['%s_TYPE' % k]

            self.save(data, data_type, k, self._count, index)

    self._count += 1

  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val


class EmptyRecorderNode(Node):
  def __init__(self):
    self._dump_dir = ''
    setattr(self, 'model_fn', None)

  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val

  def record(self, val, **kwargs):
    pass


class LocalRecorderNodeV2(object):
  def __init__(self, callback=None):
    super(LocalRecorderNodeV2, self).__init__()
    self._dump_dir = None
    self._count = 0
    self._callback = callback
    self._record_writer = None
    setattr(self, 'model_fn', None)

  def close(self):
    if self._record_writer is not None:
      self._record_writer.close()

  def save(self, data, data_type, name, count, index, params={}):
    # only support 'FILE', 'JSON', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'
    assert(data_type in ['FILE', 'JSON', 'SCALAR', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'])

    if data_type == 'FILE':
      if not os.path.exists(data):
        return None

      norm_path = os.path.normpath(data)
      file_name = '%d_%d_%s_%s'%(count, index,name,norm_path.split('/')[-1])
      shutil.copy(data, os.path.join(self.dump_dir, file_name))
      return file_name
    elif data_type == 'JSON':
      file_name = '%d_%d_%s.json'%(count, index,name)
      with open(os.path.join(self.dump_dir, file_name),'w') as fp:
        fp.write(json.dumps(data))
      return file_name
    elif data_type == 'STRING' or data_type == 'SCALAR':
      # file_name = '%d_%d_%s.txt'%(count, index,name)
      # with open(os.path.join(self.dump_dir, file_name), 'w') as fp:
      #   fp.write(str(data))
      return str(data)
    elif data_type == 'IMAGE':
      if data.dtype == np.uint8:
        transfer_result = data
      else:
        data_min = np.min(data)
        data_max = np.max(data)
        transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)

      if len(data.shape) == 3:
        assert(data.shape[2] == 3 or data.shape[2] == 4)

      assert(len(data.shape) == 2 or len(data.shape) == 3)
      file_name = '%d_%d_%s.png'%(count, index, name)
      # scipy.misc.imsave(os.path.join(self.dump_dir, file_name), transfer_result)
      # imageio.imwrite(os.path.join(self.dump_dir, file_name), transfer_result)
      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)
      cv2.imwrite(os.path.join(self.dump_dir, file_name), transfer_result)
      return file_name
    elif data_type == 'VIDEO':
      file_name = '%d_%d_%s.mp4' % (count, index, name)
      video_path = os.path.join(self.dump_dir, file_name)

      fps = 30
      if 'fps' in params:
        fps = params['fps']
      # fourcc = cv2.VideoWriter_fourcc('m', 'p', '4', 'v')
      # fourcc = cv2.VideoWriter_fourcc(*'XVID')
      fourcc = cv2.VideoWriter_fourcc(*'H264')
      vout_1 = cv2.VideoWriter(video_path, fourcc, fps, (data[0].shape[1], data[0].shape[0]))
      for frame in data:
        if frame.dtype != np.uint8:
          data_min = np.min(frame)
          data_max = np.max(frame)
          frame = ((frame - data_min) / (data_max - data_min) * 255).astype(np.uint8)
        if len(frame.shape) == 3:
          assert (frame.shape[2] == 3 or frame.shape[2] == 4)
        if len(frame.shape) == 2:
          frame = frame[:, :, np.newaxis].repeat(3, axis=2)
        assert(len(frame.shape) == 2 or len(frame.shape) == 3)

        vout_1.write(frame)
      vout_1.release()

      return file_name
    elif data_type == 'AUDIO':
      logger.error('Now, dont support save audio.')

    return None

  def record(self, val, **kwargs):
    results = val
    if type(val) != list and type(val) != tuple:
      results = [val]

    for index, result in enumerate(results):
      # 均转换成文件或字符串形式输出 [{'TYPE': 'FILE', 'PATH': ''},
      #                           {'TYPE': 'JSON', 'CONTENT': ''},
      #                           {'TYPE': 'STRING', 'CONTENT': ''},
      #                           {'TYPE': 'IMAGE', 'PATH': ''},
      #                           {'TYPE': 'VIDEO', 'PATH': ''},
      #                           {'TYPE': 'AUDIO', 'PATH': ''}]

      # 重新组织数据格式
      data = {}
      group = []
      assert('id' in result or 'image_file' in result)
      id = ''
      if 'id' in result:
        id = result['id']
        result.pop('id')
      if 'image_file' in result:
        id = result['image_file']
        result.pop('image_file')

      for key, value in result.items():
        data_name = key
        if data_name not in data:
          data[data_name] = {}
          data[data_name]['attribute'] = {}
                
        if 'type' not in value and 'TYPE' not in value:
          data[data_name]['type'] = "UNKOWN"

        if 'type' in value or 'TYPE' in value:
          if 'type' in value:
            data[data_name]['type'] = value['type']
            value.pop('type')
          else:
            data[data_name]['type'] = value['TYPE']
            value.pop('TYPE')

        data[data_name]['title'] = data_name
        if 'data' in value or 'DATA' in value:
          if 'data' in value:
            data[data_name]['data'] = value['data']
            value.pop('data')
          else:
            data[data_name]['data'] = value['DATA']
            value.pop('DATA')

        # if 'id' in value or 'ID' in value:
        #   if 'id' in value:
        #     data[data_name]['id'] = value['id']
        #     value.pop('id')
        #   else:
        #     data[data_name]['id'] = value['ID']
        #     value.pop('ID')

        # 将value中的其它作为params
        data[data_name]['params'] = value

      for key, value in data.items():
        data[key]['id'] = id

      # 转换数据格式（记录到文件）
      for data_name, data_content in data.items():
        if data[data_name]['type'] not in ['FILE', 'JSON', 'SCALAR', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO']:
          logger.error('type only support: '+','.join(['FILE', 'JSON', 'SCALAR', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO']))
          continue

        data_content['data'] = self.save(data_content['data'],
                             data_content['type'],
                             data_content['title'],
                             self._count,
                             index, 
                             data_content['params'])

        group.append(data_content)

      # 回调,后续处理
      if self._callback is not None:
        self._callback([group])
      
      # 针对保留字段，进行特殊处理
      if self._record_writer is not None:
        if 'PREDICT' in data and 'GT' in data:
          self._record_writer.write(Sample(groundtruth=data['GT']['data'], predict=data['PREDICT']['data']))

    self._count += 1

  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = val


class ActiveLearningRecorderNode(Node):
  def __init__(self, inputs):
    super(ActiveLearningRecorderNode, self).__init__(name=None,
                                                     action=self.action,
                                                     inputs=inputs,
                                                     auto_trigger=True)
    self._annotation_cache = queue.Queue()
    self._data_set = []

  @property
  def dump_dir(self):
    return None

  def action(self, *args, **kwargs):
    a = copy.deepcopy(args[0])
    if type(a) != list:
      a = [a]

    for aa in a:
      self._annotation_cache.put(aa)

  def record(self, feature):
    binded_data = {}
    binded_data['feature'] = feature

    if self._annotation_cache.qsize() > 0:
      binded_data['data'] = self._annotation_cache.get()

    self._data_set.append(binded_data)

  def iterator_value(self):
    for data in self._data_set:
      yield data



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from ...builder import DISTILL_LOSSES



'''
Modified from https://github.com/HobbitLong/RepDistiller/tree/master/crd
'''

@DISTILL_LOSSES.register_module()
class CRD(nn.Module):
	'''
	Contrastive Representation Distillation
	https://openreview.net/pdf?id=SkgpBJrtvS
	includes two symmetric parts:
	(a) using teacher as anchor, choose positive and negatives over the student side
	(b) using student as anchor, choose positive and negatives over the teacher side
	Args:
		s_dim: the dimension of student's feature
		t_dim: the dimension of teacher's feature
		feat_dim: the dimension of the projection space
		nce_n: number of negatives paired with each positive
		nce_t: the temperature
		nce_mom: the momentum for updating the memory buffer
		n_data: the number of samples in the training set, which is the M in Eq.(19)
	'''
	def __init__(self, s_dim, t_dim, feat_dim, nce_n, nce_t, nce_mom, n_data):
		super(CRD, self).__init__()
		self.embed_s = Embed(s_dim, feat_dim)
		self.embed_t = Embed(t_dim, feat_dim)
		self.contrast = ContrastMemory(feat_dim, n_data, nce_n, nce_t, nce_mom)
		self.criterion_s = ContrastLoss(n_data)
		self.criterion_t = ContrastLoss(n_data)

	def forward(self, feat_s, feat_t, idx, sample_idx):
		feat_s = self.embed_s(feat_s)
		feat_t = self.embed_t(feat_t)
		out_s, out_t = self.contrast(feat_s, feat_t, idx, sample_idx)
		loss_s = self.criterion_s(out_s)
		loss_t = self.criterion_t(out_t)
		loss = loss_s + loss_t

		return loss


class Embed(nn.Module):
	def __init__(self, in_dim, out_dim):
		super(Embed, self).__init__()
		self.linear = nn.Linear(in_dim, out_dim)

	def forward(self, x):
		x = x.view(x.size(0), -1)
		x = self.linear(x)
		x = F.normalize(x, p=2, dim=1)

		return x


class ContrastLoss(nn.Module):
	'''
	contrastive loss, corresponding to Eq.(18)
	'''
	def __init__(self, n_data, eps=1e-7):
		super(ContrastLoss, self).__init__()
		self.n_data = n_data
		self.eps = eps

	def forward(self, x):
		bs = x.size(0)
		N  = x.size(1) - 1
		M  = float(self.n_data)

		# loss for positive pair
		pos_pair = x.select(1, 0)
		log_pos  = torch.div(pos_pair, pos_pair.add(N / M + self.eps)).log_()

		# loss for negative pair
		neg_pair = x.narrow(1, 1, N)
		log_neg  = torch.div(neg_pair.clone().fill_(N / M), neg_pair.add(N / M + self.eps)).log_()

		loss = -(log_pos.sum() + log_neg.sum()) / bs

		return loss


class ContrastMemory(nn.Module):
	def __init__(self, feat_dim, n_data, nce_n, nce_t, nce_mom):
		super(ContrastMemory, self).__init__()
		self.N = nce_n
		self.T = nce_t
		self.momentum = nce_mom
		self.Z_t = None
		self.Z_s = None

		stdv = 1. / math.sqrt(feat_dim / 3.)
		self.register_buffer('memory_t', torch.rand(n_data, feat_dim).mul_(2 * stdv).add_(-stdv))
		self.register_buffer('memory_s', torch.rand(n_data, feat_dim).mul_(2 * stdv).add_(-stdv))

	def forward(self, feat_s, feat_t, idx, sample_idx):
		bs = feat_s.size(0)
		feat_dim = self.memory_s.size(1)
		n_data = self.memory_s.size(0)

		# using teacher as anchor
		weight_s = torch.index_select(self.memory_s, 0, sample_idx.view(-1)).detach()
		weight_s = weight_s.view(bs, self.N + 1, feat_dim)
		out_t = torch.bmm(weight_s, feat_t.view(bs, feat_dim, 1))
		out_t = torch.exp(torch.div(out_t, self.T)).squeeze().contiguous()

		# using student as anchor
		weight_t = torch.index_select(self.memory_t, 0, sample_idx.view(-1)).detach()
		weight_t = weight_t.view(bs, self.N + 1, feat_dim)
		out_s = torch.bmm(weight_t, feat_s.view(bs, feat_dim, 1))
		out_s = torch.exp(torch.div(out_s, self.T)).squeeze().contiguous()

		# set Z if haven't been set yet
		if self.Z_t is None:
			self.Z_t = (out_t.mean() * n_data).detach().item()
		if self.Z_s is None:
			self.Z_s = (out_s.mean() * n_data).detach().item()

		out_t = torch.div(out_t, self.Z_t)
		out_s = torch.div(out_s, self.Z_s)

		# update memory
		with torch.no_grad():
			pos_mem_t = torch.index_select(self.memory_t, 0, idx.view(-1))
			pos_mem_t.mul_(self.momentum)
			pos_mem_t.add_(torch.mul(feat_t, 1 - self.momentum))
			pos_mem_t = F.normalize(pos_mem_t, p=2, dim=1)
			self.memory_t.index_copy_(0, idx, pos_mem_t)

			pos_mem_s = torch.index_select(self.memory_s, 0, idx.view(-1))
			pos_mem_s.mul_(self.momentum)
			pos_mem_s.add_(torch.mul(feat_s, 1 - self.momentum))
			pos_mem_s = F.normalize(pos_mem_s, p=2, dim=1)
			self.memory_s.index_copy_(0, idx, pos_mem_s)

		return out_s, out_t


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/13 12:58
# @File    : preprocess.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import cv2
import numpy as np


def paddedResize(img, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, stride=32):
  # Resize and pad image while meeting stride-multiple constraints
  shape = img.shape[:2]  # current shape [height, width]
  if isinstance(new_shape, int):
    new_shape = (new_shape, new_shape)

  # Scale ratio (new / old)
  r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
  if not scaleup:  # only scale down, do not scale up (for better test mAP)
    r = min(r, 1.0)

  # Compute padding
  ratio = r, r  # width, height ratios
  new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
  dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
  if auto:  # minimum rectangle
    dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
  elif scaleFill:  # stretch
    dw, dh = 0.0, 0.0
    new_unpad = (new_shape[1], new_shape[0])
    ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

  dw /= 2  # divide padding into 2 sides
  dh /= 2

  if shape[::-1] != new_unpad:  # resize
    img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
  top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
  left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
  img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
  return img, ratio, (dw, dh)



# Copyright (c) OpenMMLab. All rights reserved.
import io
import os
import os.path as osp
import pkgutil
import re
from tabnanny import check
import time
import warnings
from collections import OrderedDict
from importlib import import_module
from tempfile import TemporaryDirectory
from numpy import mod

import torch
import torchvision
from torch.optim import Optimizer

from ..fileio import FileClient
from ..fileio import load as load_file
from ..parallel import is_module_wrapper
from ..utils import digit_version, load_url, mkdir_or_exist
from .dist_utils import get_dist_info

ENV_MMCV_HOME = 'MMCV_HOME'
ENV_XDG_CACHE_HOME = 'XDG_CACHE_HOME'
DEFAULT_CACHE_DIR = '~/.cache'


def _get_mmcv_home():
    mmcv_home = os.path.expanduser(
        os.getenv(
            ENV_MMCV_HOME,
            os.path.join(
                os.getenv(ENV_XDG_CACHE_HOME, DEFAULT_CACHE_DIR), 'mmcv')))

    mkdir_or_exist(mmcv_home)
    return mmcv_home


def load_state_dict(module, state_dict, strict=False, logger=None):
    """Load state_dict to a module.

    This method is modified from :meth:`torch.nn.Module.load_state_dict`.
    Default value for ``strict`` is set to ``False`` and the message for
    param mismatch will be shown even if strict is False.

    Args:
        module (Module): Module that receives the state_dict.
        state_dict (OrderedDict): Weights.
        strict (bool): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``False``.
        logger (:obj:`logging.Logger`, optional): Logger to log the error
            message. If not specified, print function will be used.
    """
    unexpected_keys = []
    all_missing_keys = []
    err_msg = []

    metadata = getattr(state_dict, '_metadata', None)
    state_dict = state_dict.copy()
    if metadata is not None:
        state_dict._metadata = metadata

    # use _load_from_state_dict to enable checkpoint version control
    def load(module, prefix=''):
        # recursively check parallel module in case that the model has a
        # complicated structure, e.g., nn.Module(nn.Module(DDP))
        if is_module_wrapper(module):
            module = module.module
        local_metadata = {} if metadata is None else metadata.get(
            prefix[:-1], {})
        module._load_from_state_dict(state_dict, prefix, local_metadata, True,
                                     all_missing_keys, unexpected_keys,
                                     err_msg)
        for name, child in module._modules.items():
            if child is not None:
                load(child, prefix + name + '.')

    load(module)
    load = None  # break load->load reference cycle

    # ignore "num_batches_tracked" of BN layers
    missing_keys = [
        key for key in all_missing_keys if 'num_batches_tracked' not in key
    ]

    if unexpected_keys:
        err_msg.append('unexpected key in source '
                       f'state_dict: {", ".join(unexpected_keys)}\n')
    if missing_keys:
        err_msg.append(
            f'missing keys in source state_dict: {", ".join(missing_keys)}\n')

    rank, _ = get_dist_info()
    if len(err_msg) > 0 and rank == 0:
        err_msg.insert(
            0, 'The model and loaded state dict do not match exactly\n')
        err_msg = '\n'.join(err_msg)
        if strict:
            raise RuntimeError(err_msg)
        elif logger is not None:
            logger.warning(err_msg)
        else:
            print(err_msg)


def get_torchvision_models():
    if digit_version(torchvision.__version__) < digit_version('0.13.0a0'):
        model_urls = dict()
        # When the version of torchvision is lower than 0.13, the model url is
        # not declared in `torchvision.model.__init__.py`, so we need to
        # iterate through `torchvision.models.__path__` to get the url for each
        # model.
        for _, name, ispkg in pkgutil.walk_packages(
                torchvision.models.__path__):
            if ispkg:
                continue
            _zoo = import_module(f'torchvision.models.{name}')
            if hasattr(_zoo, 'model_urls'):
                _urls = getattr(_zoo, 'model_urls')
                model_urls.update(_urls)
    else:
        # Since torchvision bumps to v0.13, the weight loading logic,
        # model keys and model urls have been changed. Here the URLs of old
        # version is loaded to avoid breaking back compatibility. If the
        # torchvision version>=0.13.0, new URLs will be added. Users can get
        # the resnet50 checkpoint by setting 'resnet50.imagent1k_v1',
        # 'resnet50' or 'ResNet50_Weights.IMAGENET1K_V1' in the config.
        json_path = osp.join(mmcv.__path__[0],
                             'model_zoo/torchvision_0.12.json')
        model_urls = mmcv.load(json_path)
        for cls_name, cls in torchvision.models.__dict__.items():
            # The name of torchvision model weights classes ends with
            # `_Weights` such as `ResNet18_Weights`. However, some model weight
            # classes, such as `MNASNet0_75_Weights` does not have any urls in
            # torchvision 0.13.0 and cannot be iterated. Here we simply check
            # `DEFAULT` attribute to ensure the class is not empty.
            if (not cls_name.endswith('_Weights')
                    or not hasattr(cls, 'DEFAULT')):
                continue
            # Since `cls.DEFAULT` can not be accessed by iterating cls, we set
            # default urls explicitly.
            cls_key = cls_name.replace('_Weights', '').lower()
            model_urls[f'{cls_key}.default'] = cls.DEFAULT.url
            for weight_enum in cls:
                cls_key = cls_name.replace('_Weights', '').lower()
                cls_key = f'{cls_key}.{weight_enum.name.lower()}'
                model_urls[cls_key] = weight_enum.url

    return model_urls


def get_external_models():
    mmcv_home = _get_mmcv_home()
    default_json_path = osp.join(mmcv.__path__[0], 'model_zoo/open_mmlab.json')
    default_urls = load_file(default_json_path)
    assert isinstance(default_urls, dict)
    external_json_path = osp.join(mmcv_home, 'open_mmlab.json')
    if osp.exists(external_json_path):
        external_urls = load_file(external_json_path)
        assert isinstance(external_urls, dict)
        default_urls.update(external_urls)

    return default_urls


def get_mmcls_models():
    mmcls_json_path = osp.join(mmcv.__path__[0], 'model_zoo/mmcls.json')
    mmcls_urls = load_file(mmcls_json_path)

    return mmcls_urls


def get_deprecated_model_names():
    deprecate_json_path = osp.join(mmcv.__path__[0],
                                   'model_zoo/deprecated.json')
    deprecate_urls = load_file(deprecate_json_path)
    assert isinstance(deprecate_urls, dict)

    return deprecate_urls


def _process_mmcls_checkpoint(checkpoint):
    if 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    else:
        # Some checkpoints converted from 3rd-party repo don't
        # have the "state_dict" key.
        state_dict = checkpoint
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if k.startswith('backbone.'):
            new_state_dict[k[9:]] = v
    new_checkpoint = dict(state_dict=new_state_dict)

    return new_checkpoint


class CheckpointLoader:
    """A general checkpoint loader to manage all schemes."""

    _schemes = {}

    @classmethod
    def _register_scheme(cls, prefixes, loader, force=False):
        if isinstance(prefixes, str):
            prefixes = [prefixes]
        else:
            assert isinstance(prefixes, (list, tuple))
        for prefix in prefixes:
            if (prefix not in cls._schemes) or force:
                cls._schemes[prefix] = loader
            else:
                raise KeyError(
                    f'{prefix} is already registered as a loader backend, '
                    'add "force=True" if you want to override it')
        # sort, longer prefixes take priority
        cls._schemes = OrderedDict(
            sorted(cls._schemes.items(), key=lambda t: t[0], reverse=True))

    @classmethod
    def register_scheme(cls, prefixes, loader=None, force=False):
        """Register a loader to CheckpointLoader.

        This method can be used as a normal class method or a decorator.

        Args:
            prefixes (str or list[str] or tuple[str]):
            The prefix of the registered loader.
            loader (function, optional): The loader function to be registered.
                When this method is used as a decorator, loader is None.
                Defaults to None.
            force (bool, optional): Whether to override the loader
                if the prefix has already been registered. Defaults to False.
        """

        if loader is not None:
            cls._register_scheme(prefixes, loader, force=force)
            return

        def _register(loader_cls):
            cls._register_scheme(prefixes, loader_cls, force=force)
            return loader_cls

        return _register

    @classmethod
    def _get_checkpoint_loader(cls, path):
        """Finds a loader that supports the given path. Falls back to the local
        loader if no other loader is found.

        Args:
            path (str): checkpoint path

        Returns:
            callable: checkpoint loader
        """
        for p in cls._schemes:
            # use regular match to handle some cases that where the prefix of
            # loader has a prefix. For example, both 's3://path' and
            # 'open-mmlab:s3://path' should return `load_from_ceph`
            if re.match(p, path) is not None:
                return cls._schemes[p]

    @classmethod
    def load_checkpoint(cls, filename, map_location=None, logger=None):
        """load checkpoint through URL scheme path.

        Args:
            filename (str): checkpoint file name with given prefix
            map_location (str, optional): Same as :func:`torch.load`.
                Default: None
            logger (:mod:`logging.Logger`, optional): The logger for message.
                Default: None

        Returns:
            dict or OrderedDict: The loaded checkpoint.
        """

        checkpoint_loader = cls._get_checkpoint_loader(filename)
        class_name = checkpoint_loader.__name__
        print(f'load checkpoint from {class_name[10:]} path: {filename}')
        return checkpoint_loader(filename, map_location)


@CheckpointLoader.register_scheme(prefixes='')
def load_from_local(filename, map_location):
    """load checkpoint by local file path.

    Args:
        filename (str): local checkpoint file path
        map_location (str, optional): Same as :func:`torch.load`.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    filename = osp.expanduser(filename)
    if not osp.isfile(filename):
        raise FileNotFoundError(f'{filename} can not be found.')
    checkpoint = torch.load(filename, map_location=map_location)
    return checkpoint


@CheckpointLoader.register_scheme(prefixes=('http://', 'https://'))
def load_from_http(filename, map_location=None, model_dir=None):
    """load checkpoint through HTTP or HTTPS scheme path. In distributed
    setting, this function only download checkpoint at local rank 0.

    Args:
        filename (str): checkpoint file path with modelzoo or
            torchvision prefix
        map_location (str, optional): Same as :func:`torch.load`.
        model_dir (string, optional): directory in which to save the object,
            Default: None

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    rank, world_size = get_dist_info()
    if rank == 0:
        checkpoint = load_url(
            filename, model_dir=model_dir, map_location=map_location)
    if world_size > 1:
        torch.distributed.barrier()
        if rank > 0:
            checkpoint = load_url(
                filename, model_dir=model_dir, map_location=map_location)
    return checkpoint


@CheckpointLoader.register_scheme(prefixes='pavi://')
def load_from_pavi(filename, map_location=None):
    """load checkpoint through the file path prefixed with pavi. In distributed
    setting, this function download ckpt at all ranks to different temporary
    directories.

    Args:
        filename (str): checkpoint file path with pavi prefix
        map_location (str, optional): Same as :func:`torch.load`.
          Default: None

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    assert filename.startswith('pavi://'), \
        f'Expected filename startswith `pavi://`, but get {filename}'
    model_path = filename[7:]

    try:
        from pavi import modelcloud
    except ImportError:
        raise ImportError(
            'Please install pavi to load checkpoint from modelcloud.')

    model = modelcloud.get(model_path)
    with TemporaryDirectory() as tmp_dir:
        downloaded_file = osp.join(tmp_dir, model.name)
        model.download(downloaded_file)
        checkpoint = torch.load(downloaded_file, map_location=map_location)
    return checkpoint


@CheckpointLoader.register_scheme(prefixes=r'(\S+\:)?s3://')
def load_from_ceph(filename, map_location=None, backend='petrel'):
    """load checkpoint through the file path prefixed with s3.  In distributed
    setting, this function download ckpt at all ranks to different temporary
    directories.

    Note:
        Since v1.4.1, the registered scheme prefixes have been enhanced to
        support bucket names in the path prefix, e.g. 's3://xx.xx/xx.path',
        'bucket1:s3://xx.xx/xx.path'.

    Args:
        filename (str): checkpoint file path with s3 prefix
        map_location (str, optional): Same as :func:`torch.load`.
        backend (str, optional): The storage backend type. Options are 'ceph',
            'petrel'. Default: 'petrel'.

    .. warning::
        :class:`mmcv.fileio.file_client.CephBackend` will be deprecated,
        please use :class:`mmcv.fileio.file_client.PetrelBackend` instead.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    allowed_backends = ['ceph', 'petrel']
    if backend not in allowed_backends:
        raise ValueError(f'Load from Backend {backend} is not supported.')

    if backend == 'ceph':
        warnings.warn(
            'CephBackend will be deprecated, please use PetrelBackend instead',
            DeprecationWarning)

    # CephClient and PetrelBackend have the same prefix 's3://' and the latter
    # will be chosen as default. If PetrelBackend can not be instantiated
    # successfully, the CephClient will be chosen.
    try:
        file_client = FileClient(backend=backend)
    except ImportError:
        allowed_backends.remove(backend)
        file_client = FileClient(backend=allowed_backends[0])

    with io.BytesIO(file_client.get(filename)) as buffer:
        checkpoint = torch.load(buffer, map_location=map_location)
    return checkpoint


@CheckpointLoader.register_scheme(prefixes=('modelzoo://', 'torchvision://'))
def load_from_torchvision(filename, map_location=None):
    """load checkpoint through the file path prefixed with modelzoo or
    torchvision.

    Args:
        filename (str): checkpoint file path with modelzoo or
            torchvision prefix
        map_location (str, optional): Same as :func:`torch.load`.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    model_urls = get_torchvision_models()
    if filename.startswith('modelzoo://'):
        warnings.warn(
            'The URL scheme of "modelzoo://" is deprecated, please '
            'use "torchvision://" instead', DeprecationWarning)
        model_name = filename[11:]
    else:
        model_name = filename[14:]

    # Support getting model urls in the same way as torchvision
    # `ResNet50_Weights.IMAGENET1K_V1` will be mapped to
    # resnet50.imagenet1k_v1.
    model_name = model_name.lower().replace('_weights', '')
    return load_from_http(model_urls[model_name], map_location=map_location)


@CheckpointLoader.register_scheme(prefixes=('open-mmlab://', 'openmmlab://'))
def load_from_openmmlab(filename, map_location=None):
    """load checkpoint through the file path prefixed with open-mmlab or
    openmmlab.

    Args:
        filename (str): checkpoint file path with open-mmlab or
        openmmlab prefix
        map_location (str, optional): Same as :func:`torch.load`.
          Default: None

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """

    model_urls = get_external_models()
    prefix_str = 'open-mmlab://'
    if filename.startswith(prefix_str):
        model_name = filename[13:]
    else:
        model_name = filename[12:]
        prefix_str = 'openmmlab://'

    deprecated_urls = get_deprecated_model_names()
    if model_name in deprecated_urls:
        warnings.warn(
            f'{prefix_str}{model_name} is deprecated in favor '
            f'of {prefix_str}{deprecated_urls[model_name]}',
            DeprecationWarning)
        model_name = deprecated_urls[model_name]
    model_url = model_urls[model_name]
    # check if is url
    if model_url.startswith(('http://', 'https://')):
        checkpoint = load_from_http(model_url, map_location=map_location)
    else:
        filename = osp.join(_get_mmcv_home(), model_url)
        if not osp.isfile(filename):
            raise FileNotFoundError(f'{filename} can not be found.')
        checkpoint = torch.load(filename, map_location=map_location)
    return checkpoint


@CheckpointLoader.register_scheme(prefixes='mmcls://')
def load_from_mmcls(filename, map_location=None):
    """load checkpoint through the file path prefixed with mmcls.

    Args:
        filename (str): checkpoint file path with mmcls prefix
        map_location (str, optional): Same as :func:`torch.load`.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """

    model_urls = get_mmcls_models()
    model_name = filename[8:]
    checkpoint = load_from_http(
        model_urls[model_name], map_location=map_location)
    checkpoint = _process_mmcls_checkpoint(checkpoint)
    return checkpoint


def _load_checkpoint(filename, map_location=None, logger=None):
    """Load checkpoint from somewhere (modelzoo, file, url).

    Args:
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``. Please refer to ``docs/model_zoo.md`` for
            details.
        map_location (str, optional): Same as :func:`torch.load`.
           Default: None.
        logger (:mod:`logging.Logger`, optional): The logger for error message.
           Default: None

    Returns:
        dict or OrderedDict: The loaded checkpoint. It can be either an
           OrderedDict storing model weights or a dict containing other
           information, which depends on the checkpoint.
    """
    return CheckpointLoader.load_checkpoint(filename, map_location, logger)


def _load_checkpoint_with_prefix(prefix, filename, map_location=None):
    """Load partial pretrained model with specific prefix.

    Args:
        prefix (str): The prefix of sub-module.
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``. Please refer to ``docs/model_zoo.md`` for
            details.
        map_location (str | None): Same as :func:`torch.load`. Default: None.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """

    checkpoint = _load_checkpoint(filename, map_location=map_location)

    if 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    else:
        state_dict = checkpoint
    if not prefix.endswith('.'):
        prefix += '.'
    prefix_len = len(prefix)

    state_dict = {
        k[prefix_len:]: v
        for k, v in state_dict.items() if k.startswith(prefix)
    }

    assert state_dict, f'{prefix} is not in the pretrained model'
    return state_dict


def load_checkpoint(model,
                    filename,
                    map_location=None,
                    strict=False,
                    logger=None,
                    revise_keys=[(r'^module\.', '')]):
    """Load checkpoint from a file or URI.

    Args:
        model (Module): Module to load checkpoint.
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``. Please refer to ``docs/model_zoo.md`` for
            details.
        map_location (str): Same as :func:`torch.load`.
        strict (bool): Whether to allow different params for the model and
            checkpoint.
        logger (:mod:`logging.Logger` or None): The logger for error message.
        revise_keys (list): A list of customized keywords to modify the
            state_dict in checkpoint. Each item is a (pattern, replacement)
            pair of the regular expression operations. Default: strip
            the prefix 'module.' by [(r'^module\\.', '')].

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    checkpoint = _load_checkpoint(filename, map_location, logger)
    # OrderedDict is a subclass of dict
    if not isinstance(checkpoint, dict):
        raise RuntimeError(
            f'No state_dict found in checkpoint file {filename}')
    # get state_dict from checkpoint
    if 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    elif 'model_state' in checkpoint:
        state_dict = checkpoint['model_state']
    else:
        state_dict = checkpoint

    # strip prefix of state_dict
    metadata = getattr(state_dict, '_metadata', OrderedDict())
    for p, r in revise_keys:
        state_dict = OrderedDict(
            {re.sub(p, r, k): v
             for k, v in state_dict.items()})
    # Keep metadata in state_dict
    state_dict._metadata = metadata

    # load state_dict
    # module = model
    # if is_module_wrapper(module):
    #     module = module.module    
    # if getattr(module, "load_state_dict", None) is not None:
    #     result = module.load_state_dict(state_dict)
    #     if result:
    #         return checkpoint

    load_state_dict(model, state_dict, strict, logger)
    return checkpoint


def weights_to_cpu(state_dict):
    """Copy a model state_dict to cpu.

    Args:
        state_dict (OrderedDict): Model weights on GPU.

    Returns:
        OrderedDict: Model weights on GPU.
    """
    state_dict_cpu = OrderedDict()
    for key, val in state_dict.items():
        state_dict_cpu[key] = val.cpu()
    # Keep metadata in state_dict
    state_dict_cpu._metadata = getattr(state_dict, '_metadata', OrderedDict())
    return state_dict_cpu


def _save_to_state_dict(module, destination, prefix, keep_vars):
    """Saves module state to `destination` dictionary.

    This method is modified from :meth:`torch.nn.Module._save_to_state_dict`.

    Args:
        module (nn.Module): The module to generate state_dict.
        destination (dict): A dict where state will be stored.
        prefix (str): The prefix for parameters and buffers used in this
            module.
    """
    for name, param in module._parameters.items():
        if param is not None:
            destination[prefix + name] = param if keep_vars else param.detach()
    for name, buf in module._buffers.items():
        # remove check of _non_persistent_buffers_set to allow nn.BatchNorm2d
        if buf is not None:
            destination[prefix + name] = buf if keep_vars else buf.detach()


def get_state_dict(module, destination=None, prefix='', keep_vars=False):
    """Returns a dictionary containing a whole state of the module.

    Both parameters and persistent buffers (e.g. running averages) are
    included. Keys are corresponding parameter and buffer names.

    This method is modified from :meth:`torch.nn.Module.state_dict` to
    recursively check parallel module in case that the model has a complicated
    structure, e.g., nn.Module(nn.Module(DDP)).

    Args:
        module (nn.Module): The module to generate state_dict.
        destination (OrderedDict): Returned dict for the state of the
            module.
        prefix (str): Prefix of the key.
        keep_vars (bool): Whether to keep the variable property of the
            parameters. Default: False.

    Returns:
        dict: A dictionary containing a whole state of the module.
    """
    # recursively check parallel module in case that the model has a
    # complicated structure, e.g., nn.Module(nn.Module(DDP))
    if is_module_wrapper(module):
        module = module.module

    # below is the same as torch.nn.Module.state_dict()
    if destination is None:
        destination = OrderedDict()
        destination._metadata = OrderedDict()
    destination._metadata[prefix[:-1]] = local_metadata = dict(
        version=module._version)
    _save_to_state_dict(module, destination, prefix, keep_vars)
    for name, child in module._modules.items():
        if child is not None:
            get_state_dict(
                child, destination, prefix + name + '.', keep_vars=keep_vars)
    for hook in module._state_dict_hooks.values():
        hook_result = hook(module, destination, prefix, local_metadata)
        if hook_result is not None:
            destination = hook_result
    return destination


def save_checkpoint(model,
                    filename,
                    optimizer=None,
                    meta=None,
                    file_client_args=None):
    """Save checkpoint to file.

    The checkpoint will have 3 fields: ``meta``, ``state_dict`` and
    ``optimizer``. By default ``meta`` will contain version and time info.

    Args:
        model (Module): Module whose params are to be saved.
        filename (str): Checkpoint filename.
        optimizer (:obj:`Optimizer`, optional): Optimizer to be saved.
        meta (dict, optional): Metadata to be saved in checkpoint.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Default: None.
            `New in version 1.3.16.`
    """
    if meta is None:
        meta = {}
    elif not isinstance(meta, dict):
        raise TypeError(f'meta must be a dict or None, but got {type(meta)}')
    # meta.update(antgo_version=antgo.__version__, time=time.asctime())

    if is_module_wrapper(model):
        model = model.module

    if hasattr(model, 'CLASSES') and model.CLASSES is not None:
        # save class name to the meta
        meta.update(CLASSES=model.CLASSES)

    checkpoint = {
        'meta': meta,
        'state_dict': weights_to_cpu(get_state_dict(model))
    }
    # save optimizer state dict in the checkpoint
    if isinstance(optimizer, Optimizer):
        checkpoint['optimizer'] = optimizer.state_dict()
    elif isinstance(optimizer, dict):
        checkpoint['optimizer'] = {}
        for name, optim in optimizer.items():
            checkpoint['optimizer'][name] = optim.state_dict()

    if filename.startswith('hdfs'):
        file_client = FileClient.infer_client({'backend': 'hdfs'}, filename)
        with io.BytesIO() as f:
            torch.save(checkpoint, f)
            file_client.put(f.getvalue(), filename)
    else:
        file_client = FileClient.infer_client(file_client_args, filename)
        with io.BytesIO() as f:
            torch.save(checkpoint, f)
            file_client.put(f.getvalue(), filename)



# Copyright (c) OpenMMLab. All rights reserved.
import torch


class _BatchNormXd(torch.nn.modules.batchnorm._BatchNorm):
    """A general BatchNorm layer without input dimension check.

    Reproduced from @kapily's work:
    (https://github.com/pytorch/pytorch/issues/41081#issuecomment-783961547)
    The only difference between BatchNorm1d, BatchNorm2d, BatchNorm3d, etc
    is `_check_input_dim` that is designed for tensor sanity checks.
    The check has been bypassed in this class for the convenience of converting
    SyncBatchNorm.
    """

    def _check_input_dim(self, input):
        return


def revert_sync_batchnorm(module):
    """Helper function to convert all `SyncBatchNorm` (SyncBN) and
    `mmcv.ops.sync_bn.SyncBatchNorm`(MMSyncBN) layers in the model to
    `BatchNormXd` layers.

    Adapted from @kapily's work:
    (https://github.com/pytorch/pytorch/issues/41081#issuecomment-783961547)

    Args:
        module (nn.Module): The module containing `SyncBatchNorm` layers.

    Returns:
        module_output: The converted module with `BatchNormXd` layers.
    """
    module_output = module
    module_checklist = [torch.nn.modules.batchnorm.SyncBatchNorm]
    if hasattr(mmcv, 'ops'):
        module_checklist.append(mmcv.ops.SyncBatchNorm)
    if isinstance(module, tuple(module_checklist)):
        module_output = _BatchNormXd(module.num_features, module.eps,
                                     module.momentum, module.affine,
                                     module.track_running_stats)
        if module.affine:
            # no_grad() may not be needed here but
            # just to be consistent with `convert_sync_batchnorm()`
            with torch.no_grad():
                module_output.weight = module.weight
                module_output.bias = module.bias
        module_output.running_mean = module.running_mean
        module_output.running_var = module.running_var
        module_output.num_batches_tracked = module.num_batches_tracked
        module_output.training = module.training
        # qconfig exists in quantized models
        if hasattr(module, 'qconfig'):
            module_output.qconfig = module.qconfig
    for name, child in module.named_children():
        module_output.add_module(name, revert_sync_batchnorm(child))
    del module
    return module_output



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 22:43
# @File    : serve.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import queue
import threading
import concurrent.futures
from antgo.pipeline.functional.entity import Entity
from antgo.pipeline.functional.option import Some


class _APIWrapper:
    """
    API Wrapper
    """
    tls = threading.local()

    def __init__(self, index=None, cls=None) -> None:
        self._queue = queue.Queue()
        self._cls = cls

        if index is not None:
            self._index = index if isinstance(index, list) else [index]
        else:
            self._index = index

    def feed(self, x) -> None:
        if self._index is None:
            entity = x
        else:
            index = self._index
            if len(index) == 2:
                input_selection, _ = index
            else:
                input_selection = index

            if type(input_selection) == str:
                input_selection = [input_selection]

            if len(input_selection) == 1:
                x = (x, )

            data = dict(zip(input_selection, x))
            entity = Entity(**data)
        entity = Some(entity)
        self._queue.put(entity)

    def __iter__(self):
        while True:
            yield self._queue.get()

    def __enter__(self):
        _APIWrapper.tls.place_holder = self
        return self._cls(self).stream()

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(_APIWrapper.tls, 'place_holder'):
            _APIWrapper.tls.place_holder = None


class _PipeWrapper:
    """
    Wrapper for execute pipeline as function
    """

    def __init__(self, pipe, place_holder) -> None:
        self._pipe = pipe
        self._place_holder = place_holder
        self._futures = queue.Queue()
        self._lock = threading.Lock()
        self._executor = threading.Thread(target=self.worker, daemon=True)
        self._executor.start()

    def worker(self):
        while True:
            future = self._futures.get()
            result = next(self._pipe)
            future.set_result(result)

    def execute(self, x):
        with self._lock:
            future = concurrent.futures.Future()
            self._futures.put(future)
            self._place_holder.feed(x)
        return future.result()


async def _decode_content(req):
    from multipart.multipart import parse_options_header
    content_type_header = req.headers.get('Content-Type')
    content_type, _ = parse_options_header(content_type_header)
    if content_type in {b'multipart/form-data'}:
        return await req.form()
    if content_type.startswith(b'image/'):
        return await req.body()
    return (await req.body()).decode()


class ServeMixin:
    """
    Mixin for API serve
    """

    def serve(self, path='/', app=None):
        """
        Serve the DataFrame as a RESTful API

        Args:
            path (str, optional): API path. Defaults to '/'.
            app (_type_, optional): The FastAPI app the API bind to, will create one if None.

        Returns:
            _type_: the app that bind to
        """
        if app is None:
            from fastapi import FastAPI, Request
            app = FastAPI()
        else:
            from fastapi import Request

        api = _APIWrapper.tls.place_holder

        pipeline = _PipeWrapper(self._iterable, api)

        @app.post(path)
        async def wrapper(req: Request):
            nonlocal pipeline
            req = await _decode_content(req)
            rsp = pipeline.execute(req)
            if rsp.is_empty():
                return rsp.get()
            return rsp.get()

        return app

    @classmethod
    def api(cls, index=None):
        return _APIWrapper(index=index, cls=cls)


from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from asyncio.log import logger

import os
import os.path as osp
import math
import time
import glob
import abc
from torch.utils.data import DataLoader
import torch.optim
import torchvision.transforms as transforms
from torch.nn.parallel.data_parallel import DataParallel
from antgo.framework.helper.utils.config import Config
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import (build_dataloader,build_kv_dataloader, build_dataset, build_iter_dataloader)
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.utils import get_logger
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.runner.builder import *
from antgo.framework.helper.models.builder import *
import torch.distributed as dist
from contextlib import contextmanager
from antgo.framework.helper.utils.setup_env import *
from antgo.framework.helper.runner.checkpoint import load_checkpoint
from antgo.framework.helper.runner.test import multi_gpu_test, single_gpu_test
from antgo.framework.helper.cnn.utils import fuse_conv_bn
from antgo.framework.helper.task_flag import *
from thop import profile
import json


class Tester(object):
    def __init__(self, cfg, work_dir='./', gpu_id=-1, distributed=False):
        if isinstance(cfg, dict):
            self.cfg = Config.fromstring(json.dumps(cfg), '.json')
        else:
            self.cfg = cfg

        device = 'cpu' if gpu_id < 0 else 'cuda'
        self.device = device
        self.distributed = distributed

        # set multi-process settings
        # 如 opencv_num_threads, OMP_NUM_THREADS, MKL_NUM_THREADS
        setup_multi_processes(self.cfg)

        self.work_dir = work_dir
        # 非分布式下，仅支持单卡
        self.cfg.gpu_ids = [gpu_id] if gpu_id >=0 else []

        if self.distributed:
            init_dist(**self.cfg.get('dist_params', {}))
            # re-set gpu_ids with distributed training mode
            _, world_size = get_dist_info()
            self.cfg.gpu_ids = range(world_size)

        if self.cfg.data.get('test', None):
            # build the dataloader
            self.dataset = build_dataset(self.cfg.data.test)
            test_dataloader_default_args = dict(
                samples_per_gpu=1, workers_per_gpu=2, dist=distributed, shuffle=False)

            test_loader_cfg = {
                **test_dataloader_default_args,
                **self.cfg.data.get('test_dataloader', {})
            }        
            
            assert(not test_loader_cfg['shuffle'])
            if getattr(self.dataset, 'is_kv', False):
                self.data_loader = build_kv_dataloader(self.dataset, **test_loader_cfg)
            elif isinstance(self.dataset, torch.utils.data.IterableDataset):
                self.data_loader = build_iter_dataloader(self.dataset, **test_loader_cfg)
            else:
                self.data_loader = build_dataloader(self.dataset, **test_loader_cfg)

    def config_model(self, model_builder=None, checkpoint='', revise_keys=[(r'^module\.', '')], is_fuse_conv_bn=False):
        # build the model and load checkpoint
        if model_builder is not None:
            self.model = model_builder()
        else:
            self.model = build_model(self.cfg.model)

        if checkpoint == '':
            checkpoint = self.cfg.get('checkpoint', checkpoint)
        
        if checkpoint is None or checkpoint == '':
            logger.error('Missing checkpoint file')
        else:
            checkpoint = load_checkpoint(self.model, checkpoint, map_location='cpu', revise_keys=revise_keys)
        
        if is_fuse_conv_bn:
            print('use fuse conv_bn')
            self.model = fuse_conv_bn(self.model)
        
        if not self.distributed:
            self.model = build_dp(self.model, self.device, device_ids=self.cfg.gpu_ids)
        else:
            print('build model in multi')
            self.model = build_ddp(
                self.model,
                self.device,
                device_ids=[int(os.environ['LOCAL_RANK'])],
                broadcast_buffers=False)

    def evaluate(self):
        # 添加运行标记
        running_flag(self.cfg.get('root', None))

        rank, _ = get_dist_info()        
        json_file = './result.json'
        if self.work_dir is not None and rank == 0:
            if not os.path.exists(osp.abspath(self.work_dir)):
                os.makedirs(osp.abspath(self.work_dir))
            timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())
            json_file = osp.join(self.work_dir, f'eval_{timestamp}.json')

        eval_cfg = self.cfg.get('evaluation', {}).copy()
        metric_func = None
        if 'metric' in eval_cfg:
            metric_func = build_measure(eval_cfg['metric'])         

        needed_info = []
        if metric_func is not None:
            needed_info = metric_func.keys()['gt']

        if not self.distributed:
            outputs = single_gpu_test(self.model, self.data_loader, needed_info=needed_info)
        else:
            outputs = multi_gpu_test(self.model, self.data_loader, needed_info=needed_info)

        if rank == 0:
            if metric_func is None:
                metric = self.dataset.evaluate(
                    outputs, **eval_cfg)
            else:
                gts = []
                for sample in outputs:
                    gt = {}
                    for k in needed_info:
                        v = sample[k]
                        gt[k] = v
                        sample.pop(k)
                    gts.append(gt)

                metric = metric_func(outputs, gts)

            print(metric)
            metric_dict = dict(metric=metric)
            if self.work_dir is not None and rank == 0:
                with open(json_file, 'w') as fp:
                    json.dump(metric_dict, fp)

        # 添加完成标记
        finish_flag(self.cfg.get('root', None))


# Copyright (c) OpenMMLab. All rights reserved.
"""This file holding some environment constant for sharing by other files."""

import os.path as osp
import subprocess
import sys
from collections import defaultdict

import cv2
import torch
from .parrots_wrapper import get_build_config


def collect_env():
    """Collect the information of the running environments.

    Returns:
        dict: The environment information. The following fields are contained.

            - sys.platform: The variable of ``sys.platform``.
            - Python: Python version.
            - CUDA available: Bool, indicating if CUDA is available.
            - GPU devices: Device type of each GPU.
            - CUDA_HOME (optional): The env var ``CUDA_HOME``.
            - NVCC (optional): NVCC version.
            - GCC: GCC version, "n/a" if GCC is not installed.
            - MSVC: Microsoft Virtual C++ Compiler version, Windows only.
            - PyTorch: PyTorch version.
            - PyTorch compiling details: The output of \
                ``torch.__config__.show()``.
            - TorchVision (optional): TorchVision version.
            - OpenCV: OpenCV version.
    """
    env_info = {}
    env_info['sys.platform'] = sys.platform
    env_info['Python'] = sys.version.replace('\n', '')

    cuda_available = torch.cuda.is_available()
    env_info['CUDA available'] = cuda_available

    if cuda_available:
        devices = defaultdict(list)
        for k in range(torch.cuda.device_count()):
            devices[torch.cuda.get_device_name(k)].append(str(k))
        for name, device_ids in devices.items():
            env_info['GPU ' + ','.join(device_ids)] = name

        from antgo.framework.helper.utils.parrots_wrapper import _get_cuda_home
        CUDA_HOME = _get_cuda_home()
        env_info['CUDA_HOME'] = CUDA_HOME

        if CUDA_HOME is not None and osp.isdir(CUDA_HOME):
            try:
                nvcc = osp.join(CUDA_HOME, 'bin/nvcc')
                nvcc = subprocess.check_output(f'"{nvcc}" -V', shell=True)
                nvcc = nvcc.decode('utf-8').strip()
                release = nvcc.rfind('Cuda compilation tools')
                build = nvcc.rfind('Build ')
                nvcc = nvcc[release:build].strip()
            except subprocess.SubprocessError:
                nvcc = 'Not Available'
            env_info['NVCC'] = nvcc

    try:
        # Check C++ Compiler.
        # For Unix-like, sysconfig has 'CC' variable like 'gcc -pthread ...',
        # indicating the compiler used, we use this to get the compiler name
        import sysconfig
        cc = sysconfig.get_config_var('CC')
        if cc:
            cc = osp.basename(cc.split()[0])
            cc_info = subprocess.check_output(f'{cc} --version', shell=True)
            env_info['GCC'] = cc_info.decode('utf-8').partition(
                '\n')[0].strip()
        else:
            # on Windows, cl.exe is not in PATH. We need to find the path.
            # distutils.ccompiler.new_compiler() returns a msvccompiler
            # object and after initialization, path to cl.exe is found.
            import locale
            import os
            from distutils.ccompiler import new_compiler
            ccompiler = new_compiler()
            ccompiler.initialize()
            cc = subprocess.check_output(
                f'{ccompiler.cc}', stderr=subprocess.STDOUT, shell=True)
            encoding = os.device_encoding(
                sys.stdout.fileno()) or locale.getpreferredencoding()
            env_info['MSVC'] = cc.decode(encoding).partition('\n')[0].strip()
            env_info['GCC'] = 'n/a'
    except subprocess.CalledProcessError:
        env_info['GCC'] = 'n/a'

    env_info['PyTorch'] = torch.__version__
    env_info['PyTorch compiling details'] = get_build_config()

    try:
        import torchvision
        env_info['TorchVision'] = torchvision.__version__
    except ModuleNotFoundError:
        pass

    env_info['OpenCV'] = cv2.__version__
    return env_info



# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : lfw.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
from antgo.dataflow.dataset import *
import os
import numpy as np
import time
import copy
import cv2
from antgo.framework.helper.fileio.file_client import *


__all__ = ['LFW']

class LFW(Dataset):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(LFW, self).__init__(train_or_test, dir, ext_params)
    assert(train_or_test in ['train', 'val', 'test'])

    if os.environ.get('LOCAL_RANK', 0) == 0:
      if not os.path.exists(os.path.join(self.dir, 'lfw-deepfunneled')):  
        # 数据集不存在，需要重新下载，并创建标记
        if not os.path.exists(self.dir):
          os.makedirs(self.dir)

        ali = AliBackend()
        ali.download('ali:///dataset/lfw/lfw-deepfunneled.tgz', self.dir)
        ali.download('ali:///dataset/lfw/pairsDevTest.txt', self.dir)
        ali.download('ali:///dataset/lfw/pairsDevTrain.txt', self.dir)
        ali.download('ali:///dataset/lfw/peopleDevTest.txt', self.dir)
        ali.download('ali:///dataset/lfw/peopleDevTrain.txt', self.dir)
        ali.download('ali:///dataset/lfw/pairs.txt', self.dir)
        ali.download('ali:///dataset/lfw/people.txt', self.dir)

        os.system(f'cd {self.dir} && tar -xf lfw-deepfunneled.tgz')
        os.system('touch DATASET_IS_READY')
      else:
        # 数据集存在，创建标记
        if not os.path.exists('DATASET_IS_READY'):
          os.system('touch DATASET_IS_READY')
    else:
      while True:
        # 等待直到存在指定文件
        if os.path.exists('DATASET_IS_READY'):
          break
        time.sleep(5)

    self.purpose = getattr(self, 'purpose', 'development')
    assert(self.purpose in ['development', 'benchmark'])

    # 1.step data folder (wild or align)
    self._data_folder = os.path.join(self.dir, 'lfw-deepfunneled')

    # 2.step data files
    self._persons_list = []
    self._persons_id_str = []
    self._persons_id = []
    for person_id_str in os.listdir(self._data_folder):
      if person_id_str[0] == '.':
        continue
      for person_image in os.listdir(os.path.join(self._data_folder, person_id_str)):
        self._persons_list.append(os.path.join(self._data_folder, person_id_str, person_image))
        self._persons_id_str.append(person_id_str)

    id_set = set(self._persons_id_str)
    self.person_id_map = {}
    for s_i, s in enumerate(id_set):
      self.person_id_map[s] = s_i

    for person_id_str in self._persons_id_str:
      person_id = self.person_id_map[person_id_str]
      self._persons_id.append(person_id)
    self.ids = list(range(len(self._persons_list)))

    self.same_pairs = []
    self.diff_pairs = []
    if self.purpose == 'development':
      # 区分训练和测试
      if self.train_or_test == 'train':
        with open(os.path.join(self.dir, 'pairsDevTrain.txt')) as fp:
          same_pair_num = fp.readline()
          same_pair_num = int(same_pair_num.strip())
          for pair_i in range(same_pair_num):
            content = fp.readline()
            person_name, image_a, image_b = content.strip().split('\t')
            person_a_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_a))
            person_b_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_b))

            self.same_pairs.append((person_a_image, person_b_image))

          content = fp.readline()
          while content:
            content = content.strip()
            if content == '':
              break
            diff_person_a_name, diff_person_a_index, diff_person_b_name, diff_person_b_index = \
              content.split('\t')

            diff_person_a_image = '%s/%s_%04d.jpg'%(diff_person_a_name, diff_person_a_name, int(diff_person_a_index))
            diff_person_b_image = '%s/%s_%04d.jpg'%(diff_person_b_name, diff_person_b_name, int(diff_person_b_index))
            self.diff_pairs.append((diff_person_a_image, diff_person_b_image))

            content = fp.readline()
      elif self.train_or_test == 'test':
        with open(os.path.join(self.dir, 'pairsDevTest.txt')) as fp:
          same_pair_num = fp.readline()
          same_pair_num = int(same_pair_num.strip())
          for pair_i in range(same_pair_num):
            content = fp.readline()
            person_name, image_a, image_b = content.strip().split('\t')
            person_a_image = '%s/%s_%04d.jpg' % (person_name, person_name, int(image_a))
            person_b_image = '%s/%s_%04d.jpg' % (person_name, person_name, int(image_b))

            self.same_pairs.append((person_a_image, person_b_image))

          content = fp.readline()
          while content:
            content = content.strip()
            if content == '':
              break
            diff_person_a_name, diff_person_a_index, diff_person_b_name, diff_person_b_index = \
              content.split('\t')

            diff_person_a_image = '%s/%s_%04d.jpg' % (diff_person_a_name, diff_person_a_name, int(diff_person_a_index))
            diff_person_b_image = '%s/%s_%04d.jpg' % (diff_person_b_name, diff_person_b_name, int(diff_person_b_index))
            self.diff_pairs.append((diff_person_a_image, diff_person_b_image))

            content = fp.readline()
    else:
      # 不区分训练和测试
      with open(os.path.join(self.dir, 'pairs.txt')) as fp:
          for line in fp.readlines()[1:]:
              pair = line.strip().split()
              if len(pair) == 3:
                # same
                person_name, image_a, image_b = pair
                person_a_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_a))
                person_b_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_b))
                self.same_pairs.append((person_a_image, person_b_image))
              else:
                # diff
                diff_person_a_name, diff_person_a_index, diff_person_b_name, diff_person_b_index = pair
                diff_person_a_image = '%s/%s_%04d.jpg' % (diff_person_a_name, diff_person_a_name, int(diff_person_a_index))
                diff_person_b_image = '%s/%s_%04d.jpg' % (diff_person_b_name, diff_person_b_name, int(diff_person_b_index))
                self.diff_pairs.append((diff_person_a_image, diff_person_b_image))

  @property
  def size(self):
    return len(self.same_pairs) + len(self.diff_pairs)

  def at(self, id):
    pair_i = id
    if pair_i < len(self.same_pairs):
      same_person_a, same_person_b = self.same_pairs[pair_i]
      same_person_a_path = os.path.join(self._data_folder, same_person_a)
      same_person_a_image = cv2.imread(same_person_a_path)
      same_person_b_path = os.path.join(self._data_folder, same_person_b)
      same_person_b_image = cv2.imread(same_person_b_path)
      
      cx = same_person_a_image.shape[1]/2
      cy = same_person_a_image.shape[0]/2
      left_t_x = int(cx - 64)
      left_t_y = int(cy - 64)
      
      right_b_x = int(cx + 64)
      right_b_y = int(cy + 64)
      
      crop_a_image = same_person_a_image[left_t_y:right_b_y, left_t_x:right_b_x]
      crop_b_image = same_person_b_image[left_t_y:right_b_y, left_t_x:right_b_x]
      return crop_a_image, {
        'issame': 1,
        'image_2': crop_b_image,
        'image_meta': {
          'image_shape': (crop_a_image.shape[0], crop_a_image.shape[1]),
          'image_file': same_person_a_path
        }
      }
    else:
      diff_person_a, diff_person_b = self.diff_pairs[pair_i-len(self.same_pairs)]
      diff_person_a_path = os.path.join(self._data_folder, diff_person_a)
      diff_person_a_image = cv2.imread(diff_person_a_path)
      diff_person_b_path = os.path.join(self._data_folder, diff_person_b)
      diff_person_b_image = cv2.imread(diff_person_b_path)
      
      cx = diff_person_a_image.shape[1]/2
      cy = diff_person_a_image.shape[0]/2
      left_t_x = int(cx - 64)
      left_t_y = int(cy - 64)
      
      right_b_x = int(cx + 64)
      right_b_y = int(cy + 64)
      
      crop_a_image = diff_person_a_image[left_t_y:right_b_y, left_t_x:right_b_x]
      crop_b_image = diff_person_b_image[left_t_y:right_b_y, left_t_x:right_b_x]      
      return crop_a_image, {
        'issame': 0,
        'image_2': crop_b_image,
        'image_meta': {
          'image_shape': (crop_a_image.shape[0], crop_a_image.shape[1]),
          'image_file': diff_person_a_path
        }
      }

# lfw = LFW('train', '/root/workspace/dataset/lfw', ext_params={'purpose': 'benchmark'})
# num = lfw.size
# for i in range(num):
#   data = lfw.sample(i)
#   # print(data.keys())
#   cv2.imwrite('./a1.png', data['image'][0])
#   cv2.imwrite('./b1.png', data['image'][1])
#   print(i)



from setuptools import setup
from setuptools import dist
from setuptools import find_packages
from distutils.extension import Extension
from antgo import __version__
from os import path as os_path
import os
import shutil


this_directory = os_path.abspath(os_path.dirname(__file__))
dist.Distribution().fetch_build_eggs(['numpy'])

# install: python setup.py build_ext install -r requirements.txt (from github)
# install: pip install .
def ext_modules():
    import numpy as np

    # some_extention = Extension(..., include_dirs=[np.get_include()])
    some_extention = [
      Extension('antgo.utils._mask',
                sources=['antgo/cutils/maskApi.c', 'antgo/utils/_mask.pyx'],
                include_dirs=[np.get_include(), 'antgo/cutils'],
                extra_compile_args=['-Wno-cpp', '-Wno-unused-function', '-std=c99'],
                ),
      Extension('antgo.utils._bbox',
                ["antgo/utils/_bbox.pyx"],
                include_dirs=[np.get_include()],
                extra_compile_args=["-Wno-cpp", "-Wno-unused-function", '-std=c99']
                )
    ]

    return some_extention


def readme():
    with open('README.rst') as f:
        return f.read()


def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]

setup(name='antgo',
      version=str(__version__),
      description='machine learning experiment platform',
      __short_description__='machine learning experiment platform',
      url='https://github.com/jianzfb/antgo',
      author='jian',
      author_email='jian@mltalker.com',
      setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        'setuptools>=18.0',
        'Cython',
      ],
      packages=find_packages(),
      ext_modules=ext_modules(),
      entry_points={'console_scripts': ['antgo=antgo.main:main'], },
      # install_requires=read_requirements('requirements.txt'),
      long_description=readme(),
      include_package_data=True,
      zip_safe=False,)



# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from paddle.fluid.core import VarBase
from antgo.framework.torch2paddle.utils import paddle_dtypes


def is_condition_one(idx):
    """
    a = paddle.to_tensor(np.array([[1,2,3], [4,5,6]]).astype("float32"))
    mask = paddle.to_tensor(np.array([True, False]).astype("bool"))
    a[mask, :]
    a[mask, ...]
    """
    if not (isinstance(idx[0], paddle.Tensor) and \
            idx[0].dtype == paddle_dtypes.t_bool):
        return False
    if len(idx) == 1:
        return False
    if len(idx) > 1:
        if idx[1] is Ellipsis:
            return True
        for ele in idx[1:]:
            if isinstance(
                    ele, slice
            ) and ele.start is None and ele.start is None and ele.step is None:
                continue
            else:
                return False
    return True


def is_condition_two(idx):
    """
    a = paddle.to_tensor(np.random.rand(1, 2, 3).astype("float32"))
    a[..., :2]
    """
    if idx[0] is Ellipsis and (isinstance(idx[1], slice) or isinstance(idx[1],
                                                                       int)):
        return True
    return False


VarBase.tmp = VarBase.__getitem__


def __getitem__(self, idx):
    is_bool = False
    if self.dtype == paddle_dtypes.t_bool:
        self = self.cast("int32")
        is_bool = True
    if isinstance(idx, paddle.Tensor) and len(idx.shape) == 1:
        out = paddle.gather(self, idx)
        return out.cast("bool") if is_bool else out
    elif isinstance(idx, paddle.Tensor) and idx.dtype == paddle_dtypes.t_bool:
        idx = paddle.cast(idx, "int32")
        idx = paddle.nonzero(idx)
        out = paddle.gather_nd(self, idx)
        return out.cast("bool") if is_bool else out
    elif isinstance(idx, tuple):
        if is_condition_one(idx):
            first_idx = idx[0]
            first_idx = paddle.cast(first_idx, "int32")
            first_idx = paddle.nonzero(first_idx)
            out = paddle.gather_nd(self, first_idx)
            return out.cast("bool") if is_bool else out
        elif is_condition_two(idx):
            new_idx = list()
            for i in range(len(self.shape) - 1):
                new_idx.append(slice(None, None, None))
            new_idx.append(list(idx)[-1])
            out = self.tmp(tuple(new_idx))
            return out.cast("bool") if is_bool else out
        else:
            out = self.tmp(idx)
            return out.cast("bool") if is_bool else out
        # TODO(syf): 出来为(slice(None, None, None), slice(None, None, None), 0)
    else:
        out = self.tmp(idx)
        if out.shape == [1]:
            return out.numpy()[0]
        else:
            return out


VarBase.__getitem__ = __getitem__

VarBase.setitem_tmp = VarBase.__setitem__


def __setitem__(self, idx, value):
    if isinstance(idx, paddle.Tensor) and idx.dtype == paddle_dtypes.t_bool:
        """
        a = paddle.to_tensor(np.array([1,2,3]).astype("float32"))
        mask = paddle.to_tensor(np.array([True, False, True]).astype("bool"))
        a[mask] = 1
        """
        value_tensor = paddle.full(self.shape, value, self.dtype)
        paddle.assign(paddle.where(idx, value_tensor, self), self)
    else:
        return self.setitem_tmp(idx, value)


VarBase.__setitem__ = __setitem__



# Copyright (c) OpenMMLab. All rights reserved.
import math

from torch import nn
from torch.nn import functional as F

from .registry import CONV_LAYERS


@CONV_LAYERS.register_module()
class Conv2dAdaptivePadding(nn.Conv2d):
    """Implementation of 2D convolution in tensorflow with `padding` as "same",
    which applies padding to input (if needed) so that input image gets fully
    covered by filter and stride you specified. For stride 1, this will ensure
    that output image size is same as input. For stride of 2, output dimensions
    will be half, for example.

    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of
            the input. Default: 0
        dilation (int or tuple, optional): Spacing between kernel elements.
            Default: 1
        groups (int, optional): Number of blocked connections from input
            channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 bias=True):
        super().__init__(in_channels, out_channels, kernel_size, stride, 0,
                         dilation, groups, bias)

    def forward(self, x):
        img_h, img_w = x.size()[-2:]
        kernel_h, kernel_w = self.weight.size()[-2:]
        stride_h, stride_w = self.stride
        output_h = math.ceil(img_h / stride_h)
        output_w = math.ceil(img_w / stride_w)
        pad_h = (
            max((output_h - 1) * self.stride[0] +
                (kernel_h - 1) * self.dilation[0] + 1 - img_h, 0))
        pad_w = (
            max((output_w - 1) * self.stride[1] +
                (kernel_w - 1) * self.dilation[1] + 1 - img_w, 0))
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [
                pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2
            ])
        return F.conv2d(x, self.weight, self.bias, self.stride, self.padding,
                        self.dilation, self.groups)



class Flags:
    def __init__(self, s=''):
        self.flags = s

    def add_definition(self, key, value):
        if isinstance(value, bool):
            value = int(value)
        self.flags += ' -D%s=%s' % (key, str(value))
        return self

    def add_string(self, s):
        self.flags += ' %s' % str(s)
        return self

    def __str__(self):
        return self.flags



import ctypes

CTYPE_INTS = [ctypes.c_short, ctypes.c_int, ctypes.c_long, ctypes.c_longlong]
CTYPE_UINTS = [ctypes.c_ushort, ctypes.c_uint,
               ctypes.c_ulong, ctypes.c_ulonglong]
CTYPENAME2CTYPE = {
    'bool': ctypes.c_bool,
    'char': ctypes.c_char,
    'char*': ctypes.c_char_p,
    'double': ctypes.c_double,
    'float': ctypes.c_float,
    'int': ctypes.c_int,
    'int8_t': ctypes.c_int8,
    'int16_t': ctypes.c_int16,
    'int32_t': ctypes.c_int32,
    'int64_t': ctypes.c_int64,
    'void*': ctypes.c_void_p,
    'void': None,
    None: None,
}


def get_ctype_name(ctype):
    # ctype.__name__ = 'c_xxx'
    if ctype in CTYPE_INTS[2:]:
        return 'int{}_t'.format(ctypes.sizeof(ctype) * 8)
    if ctype in CTYPE_UINTS[2:]:
        return 'uint{}_t'.format(ctypes.sizeof(ctype) * 8)
    name = ctype.__name__
    if name.startswith('c_'):
        name = name[2:]
    return name


class DType:
    _EXTRA_ATTRS = dict()

    def __init__(self, ctype, is_const=False):
        self.ctype = ctype
        self.is_const = is_const
        self.cname, self.is_pointer = self._get_extra_attr()

    def _get_extra_attr(self):
        idcode = (self.ctype, self.is_const)
        if idcode not in DType._EXTRA_ATTRS:
            # Assignment for self.is_pointer and self.cname
            if self.ctype.__name__.startswith('LP'):
                is_pointer = True
                basic_type = self.ctype._type_
            else:
                is_pointer = False
                basic_type = self.ctype

            ctype_name = get_ctype_name(basic_type)

            if self.is_const:
                ctype_name = 'const {}'.format(ctype_name)
            if is_pointer:
                ctype_name += '*'
            cname = ctype_name
            DType._EXTRA_ATTRS[idcode] = (cname, is_pointer)
        return DType._EXTRA_ATTRS[idcode]

    def __repr__(self):
        return self.cname

    def __call__(self, value):
        return self.ctype(value)


class CStruct:
    def __init__(self, name, is_const, cstruct, constructor):
        self.cname = name + '*'
        self.cstruct = cstruct
        self.constructor = constructor
        self.is_pointer = True
        self.is_const = is_const

    def __call__(self, *args, **kwargs):
        return self.constructor(*args, **kwargs)


class UnknownCType:
    def __init__(self, tname):
        self.tname = tname
        self.is_const = False


class TemplateType:
    def __init__(self, tname, is_pointer, is_const):
        self.tname = tname
        self.is_pointer = is_pointer
        self.is_const = is_const

    def __repr__(self):
        return '<typename {const}{tname}{pointer}>'.format(
            const='const ' if self.is_const else '',
            tname=self.tname,
            pointer='*' if self.is_pointer else '')

    def __call__(self, ctype):
        if self.is_pointer:
            ctype = ctypes.POINTER(ctype)
        return DType(ctype, is_const=self.is_const)



from antgo.framework.helper.parallel import is_module_wrapper
from antgo.framework.helper.runner.hooks import HOOKS, Hook
from bisect import bisect_right


@HOOKS.register_module()
class MeanTeacher(Hook):
    def __init__(
        self,
        momentum=0.999,
        interval=1,
        warm_up=100,
        decay_intervals=None,
        decay_factor=0.0,
    ):
        assert momentum >= 0 and momentum <= 1
        self.momentum = momentum
        assert isinstance(interval, int) and interval > 0
        self.warm_up = warm_up
        self.interval = interval
        assert isinstance(decay_intervals, list) or decay_intervals is None
        self.decay_intervals = decay_intervals
        self.decay_factor = decay_factor

    def before_run(self, runner):
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        assert hasattr(model, "teacher")
        assert hasattr(model, "student")
        # only do it at initial stage
        if runner.iter == 0:
            # log_every_n("Clone all parameters of student to teacher...")
            self.momentum_update(model, 0)

    def before_train_iter(self, runner):
        """Update ema parameter every self.interval iterations."""
        curr_step = runner.iter
        if curr_step % self.interval != 0:
            return
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        if self.decay_factor <= 1e-4:
            # keep the same
            momentum = self.momentum
        else:
            # We warm up the momentum considering the instability at beginning
            momentum = min(
                self.momentum, 1 - (1 + self.warm_up) / (curr_step + 1 + self.warm_up)
            )            

        runner.log_buffer.output["ema_momentum"] = momentum
        self.momentum_update(model, momentum)

    def after_train_iter(self, runner):
        curr_step = runner.iter
        if self.decay_intervals is None:
            return
        
        # if decay_factory == 0.0, keep the same all the way
        if self.decay_factor > 1e-4:
            self.momentum = 1 - (1 - self.momentum) * self.decay_factor ** bisect_right(
                self.decay_intervals, curr_step
            )

    def momentum_update(self, model, momentum):
        for (src_name, src_parm), (tgt_name, tgt_parm) in zip(
            model.student.named_parameters(), model.teacher.named_parameters()
        ):
            tgt_parm.data.mul_(momentum).add_(src_parm.data, alpha=1 - momentum)



# Copyright (c) OpenMMLab. All rights reserved.
from itertools import chain

from torch.nn.parallel import DataParallel

from .scatter_gather import scatter_kwargs


class MMDataParallel(DataParallel):
    """The DataParallel module that supports DataContainer.

    MMDataParallel has two main differences with PyTorch DataParallel:

    - It supports a custom type :class:`DataContainer` which allows more
      flexible control of input data during both GPU and CPU inference.
    - It implement two more APIs ``train_step()`` and ``val_step()``.

    .. warning::
        MMDataParallel only supports single GPU training, if you need to
        train with multiple GPUs, please use MMDistributedDataParallel
        instead. If you have multiple GPUs and you just want to use
        MMDataParallel, you can set the environment variable
        ``CUDA_VISIBLE_DEVICES=0`` or instantiate ``MMDataParallel`` with
        ``device_ids=[0]``.

    Args:
        module (:class:`nn.Module`): Module to be encapsulated.
        device_ids (list[int]): Device IDS of modules to be scattered to.
            Defaults to None when GPU is not available.
        output_device (str | int): Device ID for output. Defaults to None.
        dim (int): Dimension used to scatter the data. Defaults to 0.
    """

    def __init__(self, *args, dim=0, **kwargs):
        super(MMDataParallel, self).__init__(*args, dim=dim, **kwargs)
        self.dim = dim

    def forward(self, *inputs, **kwargs):
        """Override the original forward function.

        The main difference lies in the CPU inference where the data in
        :class:`DataContainers` will still be gathered.
        """
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module(*inputs[0], **kwargs[0])
        else:
            return super().forward(*inputs, **kwargs)

    def scatter(self, inputs, kwargs, device_ids):
        return scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)

    def train_step(self, *inputs, **kwargs):
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module.train_step(*inputs[0], **kwargs[0])

        assert len(self.device_ids) == 1, \
            ('MMDataParallel only supports single GPU training, if you need to'
             ' train with multiple GPUs, please use MMDistributedDataParallel'
             ' instead.')

        # for t in chain(self.module.parameters(), self.module.buffers()):
        #     if t.device != self.src_device_obj:
        #         raise RuntimeError(
        #             'module must have its parameters and buffers '
        #             f'on device {self.src_device_obj} (device_ids[0]) but '
        #             f'found one of them on device: {t.device}')

        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        return self.module.train_step(*inputs[0], **kwargs[0])

    def val_step(self, *inputs, **kwargs):
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module.val_step(*inputs[0], **kwargs[0])

        assert len(self.device_ids) == 1, \
            ('only supports single GPU training')

        # for t in chain(self.module.parameters(), self.module.buffers()):
        #     if t.device != self.src_device_obj:
        #         raise RuntimeError(
        #             'module must have its parameters and buffers '
        #             f'on device {self.src_device_obj} (device_ids[0]) but '
        #             f'found one of them on device: {t.device}')

        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        return self.module.val_step(*inputs[0], **kwargs[0])



# -*- coding: UTF-8 -*-
# @Time    : 17-8-2
# @File: dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import os
import random
import numpy as np
from antgo.utils import get_rng
from antgo.dataflow.core import *
from antgo.utils.fs import *
from functools import reduce
import re
import multiprocessing
import tarfile
import zipfile
from antgo import config
import subprocess
from antgo.utils.serialize import *
import copy
import cv2
try:
    import queue
except:
    import Queue as queue
import functools

Config = config.AntConfig


def imread(file):
  return cv2.imread(file)


def imwrite(file, img):
  cv2.imwrite(file, img)


def imresize(image, size):
  return cv2.resize(image, (size[1],size[0]))


class Dataset(BaseNode):
  _BASE_DATASET = True

  def __init__(self, train_or_test="train", dir=None, ext_params=None, name=None):
    super(Dataset, self).__init__(name)
    assert(train_or_test in ["train", "val", "test", "sample", 'unlabeled'])

    # for random episodes
    self.nb_samples = 0
    self.nb_samples_per_class = 0

    # basic info
    self.train_or_test = train_or_test
    self.dir = dir
    
    self._dataset_type = 'IMAGE'
    # config extent params
    # included, excluded, transform(cls->id)
    self.ext_params = ext_params
    if ext_params is not None:
      for k, v in ext_params.items():
        if k != 'self':
          setattr(self, k, v)

    self.data_rng = None

    self.epochs = None
    self._epoch = 0
    self.data_generator = None
    self._lock = multiprocessing.Lock()
    self._ids = []
    # data rng flag
    self._is_data_rng = False

    self._candidate_file = None
    self._unlabeled_list_file = None
    if self.dir is not None:
      self._candidate_file = os.path.join(self.dir, 'candidates.txt')
      self._unlabeled_list_file = os.path.join(self.dir, 'unlabeled_list.txt')

  def init(self, *args, **kwargs):
    pass
  
  def close(self):
    pass

  @property
  def unlabeled_tag(self):
    return 'unlabeled'

  @property
  def unlabeled_folder(self):
    return os.path.join(self.dir, self.unlabeled_tag)

  @property
  def candidate_folder(self):
    return os.path.join(self.dir, 'candidates')

  @property
  def candidate_file(self):
    return self._candidate_file
  @candidate_file.setter
  def candidate_file(self, val):
    self._candidate_file = val

  @property
  def unlabeled_list_file(self):
    return self._unlabeled_list_file
  @unlabeled_list_file.setter
  def unlabeled_list_file(self, val):
    self._unlabeled_list_file = val

  def unlabeled_size(self):
    unlabeled_folder = os.path.join(self.dir, self.unlabeled_tag)
    if not os.path.exists(unlabeled_folder):
      return 0

    if not os.path.exists(self.unlabeled_list_file):
      # 构建无标注数据文件列表
      unorder_list = []
      for file in os.listdir(unlabeled_folder):
        if file[0] == '.':
          continue
        unorder_list.append(file)
      
      order_list = sorted(unorder_list)
      with open(self.unlabeled_list_file, 'w') as fp:
        for file in order_list:
          fp.write('%s/%s,%d\n' % (self.unlabeled_tag, file, 0))

    has_labeled_list = []
    if os.path.exists(self.candidate_file):
      with open(self.candidate_file, 'r') as fp:
        line_content = fp.readline()
        while line_content:
          _, data_file, _ = line_content.split(',')
          has_labeled_list.append(data_file.split('/')[-1])
          line_content = fp.readline()
    else:
      # build empty candidates file
      with open(self.candidate_file, 'w') as fp:
        pass
    
    unlabeled_num = 0
    with open(self.unlabeled_list_file, 'r') as fp:
      line_content = fp.readline()
      while line_content:
        file_name, _ = line_content.split(',')
        if file_name.split('/')[-1] not in has_labeled_list:
          unlabeled_num += 1

        line_content = fp.readline()

    return unlabeled_num

  def unlabeled(self, tag=''):
    unlabeled_folder = os.path.join(self.dir, self.unlabeled_tag)
    if not os.path.exists(unlabeled_folder):
      return None

    if not os.path.exists(self.unlabeled_list_file):
      # 构建无标注数据文件列表
      unorder_list = []
      for file in os.listdir(unlabeled_folder):
        if file[0] == '.':
          continue
        unorder_list.append(file)
      
      order_list = sorted(unorder_list)
      with open(self.unlabeled_list_file, 'w') as fp:
        for file in order_list:
          fp.write('%s/%s,%d\n' % (self.unlabeled_tag, file, 0))

    has_labeled_list = []
    if os.path.exists(self.candidate_file):
      with open(self.candidate_file, 'r') as fp:
        line_content = fp.readline()
        while line_content:
          _, data_file, _ = line_content.split(',')
          has_labeled_list.append(data_file.split('/')[-1])
          line_content = fp.readline()
    else:
      # build empty candidates file
      with open(self.candidate_file, 'w') as fp:
        pass

    unlabeled_list = []
    unlabeled_index = 0
    with open(self.unlabeled_list_file, 'r') as fp:
      line_content = fp.readline()
      while line_content:
        file_name, _ = line_content.split(',')
        unlabeled_list.append((unlabeled_index,file_name))

        line_content = fp.readline()
        unlabeled_index += 1
    
    unlabeled_data = []
    for index, file in unlabeled_list:
      if file.split('/')[-1] not in has_labeled_list:
        unlabeled_data.append({'id': index, 'file_id': file})

    return unlabeled_data

  def make_candidate(self, unlabeled_id, unlabeled_file, label_file, status=''):
    # status: 'SKIP/OK'
    if not os.path.exists(os.path.join(self.dir, 'candidates')):
      os.makedirs(os.path.join(self.dir, 'candidates'))
      os.makedirs(os.path.join(self.dir, 'candidates', 'data'))
      os.makedirs(os.path.join(self.dir, 'candidates', 'label'))

    with open(self.candidate_file, 'a') as fp:
      if status == 'OK':
        # copy data file to candidates/data
        shutil.copy(os.path.join(self.dir, unlabeled_file), os.path.join(self.dir, 'candidates', 'data'))

        # copy label file to candidates/label
        shutil.copy(label_file, os.path.join(self.dir, 'candidates', 'label'))

        # write to canndidates.txt
        fp.write('%d,%s,%s\n'%(unlabeled_id, 'candidates/data/%s'%unlabeled_file.split('/')[-1], 'candidates/label/%s'%label_file.split('/')[-1]))

  def check_candidate(self, unlabeled_files, finished_label_folder):
    # 检查准备进入候选列表的标注数据，与等待的未标注数据一致
    consistent_sample_num = 0
    logger.info("Finshed label folder %s"%finished_label_folder)

    for sample_file, _ in unlabeled_files:
      if os.path.exists(os.path.join(finished_label_folder, sample_file)):
        consistent_sample_num += 1

    logger.info('Consistent unlabeled file and labeled file %d'%consistent_sample_num)
    if consistent_sample_num != len(unlabeled_files):
      return False

    return True

  def candidates(self, candidate_type='IMAGE'):
    '''

    :return:
    '''
    if not os.path.exists(self.candidate_file):
      raise StopIteration

    with open(self.candidate_file, 'r') as fp:
      line_content = fp.readline()
      while line_content:
        _, data_file, label_file = line_content.split(',')
        data_file = data_file.strip()
        label_file = label_file.strip()
        if candidate_type == 'IMAGE':
          yield imread(os.path.join(self.dir, data_file)), \
                imread(os.path.join(self.dir, label_file))
        else:
          yield os.path.join(self.dir, data_file), \
                os.path.join(self.dir, label_file)

        line_content = fp.readline()

    raise StopIteration

  def candidates_size(self):
    if not os.path.exists(self.candidate_file):
      return 0

    count = 0
    with open(self.candidate_file, 'r') as fp:
      line_content = fp.readline()
      while line_content:
        count += 1
        line_content = fp.readline()

    return count

  def waiting_process_num(self):
    return 0

  def finish_process_num(self):
    return 0

  @property
  def ids(self):
    return self._ids
  @ids.setter
  def ids(self, val):
    self._ids = val

  @property
  def epochs(self):
    return self._epochs
  @epochs.setter
  def epochs(self, val):
    self._epochs = val
    
  @property
  def epoch(self):
    return self._epoch
  @epoch.setter
  def epoch(self, val):
    self._epoch = val
  
  @property
  def dataset_type(self):
    return self._dataset_type
  @dataset_type.setter
  def dataset_type(self, val):
    self._dataset_type = val
  
  @property
  def data_generator(self):
    return self._data_generator
  @data_generator.setter
  def data_generator(self, val):
    self._data_generator = val

  def _reset_iteration_state(self):
    self.data_generator = None
  
  def set_value(self, new_value):
    pass
  
  def get_value(self):
    try:
      if DIRTY == self._value:
        if self.data_generator is None:
          self.data_generator = self.data_pool()
        self._value = next(self.data_generator)
        #self._set_dependents_dirty()

      return self._value
    except:
      raise StopIteration
  
  def _force_inputs_dirty(self):
    self._value = DIRTY
  
  value = property(get_value, set_value)

  def iterator_value(self):
    while True:
      self._value = self.get_value()
      yield self._value
      self._force_inputs_dirty()

  def __iter__(self):
    return self.iterator_value()

  def __len__(self):
    return self.size

  def __getitem__(self, index):
    return self.at(index)

  def at(self, id):
    raise NotImplementedError

  def get_ann_info(self, idx=None):
    raise NotImplementedError

  def get_cat_ids(self, idx):
    raise NotImplementedError
  
  def sample(self, id):
    image, annotation = self.at(id)
    if image is None:
      # None 跳过
      return None
    
    if not isinstance(annotation, dict):
      annotation = {'label': annotation}

    if image is not None:
      annotation['image'] = image
    return annotation

  def evaluate(self, preds, **kwargs):
    raise NotImplementedError

  @property
  def size(self):
    raise NotImplementedError

  @property
  def rng(self):
    if self._is_data_rng:
      self.data_rng = get_rng(self)
    return self.data_rng

  def reset_state(self):
    """
    Reset state of the dataflow. Will always be called before consuming data points.
    for example, RNG **HAS** to be reset here if used in the DataFlow.
    Otherwise it may not work well with prefetching, because different
    processes will have the same RNG state.
    """
    self._is_data_rng = True

  def split(self, split_params, split_method):
    pass

  def _split(self, idx=[], split_params={}, split_method='holdout'):
    '''
    
    :param idx:
    :param split_params:
    :param split_method:
    :return:
    '''
    if split_method == 'holdout':
      if 'ratio' not in split_params:
        split_params['ratio'] = None
      if 'is_stratified' not in split_params:
        split_params['is_stratified'] = True

      return self._split_holdout(split_params['ratio'], split_params['is_stratified'], idx)
    elif split_method == 'repeated-holdout':
      return self._split_repeated_holdout(split_params['ratio'], split_params['is_stratified'], idx)
    elif split_method == 'bootstrap':
      return self._split_bootstrap(idx)
    elif split_method == 'kfold':
      return self._split_kfold(split_params['kfold'], split_params['k'], idx)

    return None

  def _split_custom_holdout(self, split_ratio, is_stratified_sampling,idx):
    # implemented at child class
    # train/val
    raise NotImplementedError()

  def _split_holdout(self, split_ratio, is_stratified_sampling=True, idx=[]):
    assert(self.train_or_test in ['train','sample'])
    try:
      train_idx, val_idx = self._split_custom_holdout(split_ratio, is_stratified_sampling, idx)
      return train_idx, val_idx
    except:
      return self._split_repeated_holdout(split_ratio, is_stratified_sampling, idx)

  def _split_repeated_holdout(self, split_ratio, is_stratified_sampling=True, idx=[]):
    assert(self.train_or_test in ['train','sample'])

    # split by t/v
    if is_stratified_sampling:
      labels_num = len(set(idx))
      labels_index = [[] for _ in range(labels_num)]
      for index, label in enumerate(idx):
        labels_index[label].append(index)

      train_idx = [[] for _ in range(labels_num)]
      validation_idx = [[] for _ in range(labels_num)]
      for label in range(labels_num):
        label_samples = labels_index[label]
        get_rng().shuffle(label_samples)

        top_k = int(split_ratio * len(label_samples))
        train_idx[label].extend(label_samples[0:top_k])
        validation_idx[label].extend((label_samples[top_k:]))

      train_idx = reduce(lambda x, y: x+y, train_idx)
      validation_idx = reduce(lambda x, y: x+y, validation_idx)

      return train_idx, validation_idx
    else:
      index = range(len(idx))
      get_rng().shuffle(index)
      top_k = int(split_ratio * len(idx))
      train_idx = index[0:top_k]
      validation_idx = (index[top_k:])
      return train_idx, validation_idx

  def _split_bootstrap(self, idx):
    assert(self.train_or_test in ['train','sample'])

    is_ok = False
    train_idx = []
    validation_idx = []
    while not is_ok:
      selected_idx = get_rng().randint(low=0, high=len(idx)-1, size=len(idx)).tolist()
      train_idx = selected_idx
      validation_idx = [i for i in range(len(idx)) if i not in selected_idx]

      if len(validation_idx) > 0:
        is_ok = True

    return train_idx, validation_idx

  def _split_kfold(self, k_fold, k, idx):
    assert(self.train_or_test in ['train','sample'])
    assert(k < k_fold)

    size = len(idx)
    fold_size = int(float(size) / float(k_fold))
    k_start = k * fold_size
    k_end = (k+1) * fold_size if k < k_fold - 1 else len(idx)

    validation_idx = [idx[i] for i in np.arange(k_start, k_end, dtype=np.uint64)]
    train_idx = [i for i in range(len(idx)) if i not in validation_idx]

    return train_idx, validation_idx

  def data_pool(self):
    '''
    
    :return:
    '''
    raise NotImplementedError

  def make_data(self, data, label=None, support=None,transform_func=None):
    # data: [[],[],[],...]
    # label:[0,1,2,...]
    # support: [(0,[]),[1,[]],...]
    # return: [[label_index,label,data]]
    if label == None:
      label = range(len(data))

    if support == None:
      samples = [(i_index, i_val, d) for i_index, i_val in enumerate(label) for d in data[i_index]]
      if self.rng != None:
        self.rng.shuffle(samples)
      return samples
    else:
      # building samples by support
      support_labels, support_datas = zip(*support)
      table = {}
      for l_i, l_v in enumerate(support_labels):
        table[l_v] = l_i

      samples = None
      if transform_func is None:
        samples = [(table[i_val], i_val, d) for i_index, i_val in enumerate(label) \
                   if i_val in support_labels \
                   for d_index, d in enumerate(data[i_val]) \
                   if d_index not in support_datas[table[i_val]]]
      else:
        samples = [(table[i_val], i_val, d) for i_index, i_val in enumerate(label) \
                   if i_val in support_labels \
                   for d in data[i_val] \
                   if transform_func(d) not in support_datas[table[i_val]]]

      if self.rng != None:
        self.rng.shuffle(samples)
      return samples

  def make_sequence_data(self, data, label=None, nb_samples=0, nb_samples_per_class=0):
    '''
    only for image sequence
    :param data: [[],[],[],...]
    :param label: [0,1,2,...]
    :param nb_samples: sampling classes number
    :param nb_samples_per_class: sampling examples from per class
    :return:
    '''
    if label == None:
      label = range(len(data))

    label_num = len(label)

    # sample classes
    data_and_labels = zip(data, label)
    data_and_labels = [term for term in data_and_labels if len(term[0]) > 0]

    classes_samples = random.sample(data_and_labels, nb_samples)

    # sample example from every class
    sampler = lambda x: random.sample(x, nb_samples_per_class)
    samples = [(i, classes_samples[i][1], choice, classes_samples[i][0][choice]) \
               for i in range(nb_samples) \
               for choice in sampler(range(len(classes_samples[i][0])))]

    # shuffle
    if self.rng != None:
      self.rng.shuffle(samples)

    return samples

  def set_sequence(self, nb_samples, nb_samples_per_class):
    self.nb_samples = nb_samples
    self.nb_samples_per_class = nb_samples_per_class
    if nb_samples * nb_samples_per_class > 0:
      self.is_sequence = True

  def load_image(self,image_path):
    return imread(image_path)

  def reorganize_data(self,data,label,class_num=0):
    '''

    Parameters
    ----------
    data list (image list)
    label list (label list)

    Returns
    -------
        [[image,index],[],...]
    '''
    if class_num == 0:
      class_num = len(set(label))

    data_group = [[] for i in range(class_num)]
    for index in range(len(data)):
      data_group[label[index]].append((data[index],index))

    return data_group

  def regorganize_object_data(self,data,data_label,info):
    '''

    Parameters
    ----------
    data list (image list)
    data_label list (image label list)
    info class (dict)

    Returns
    -------
        [[data,data_index,(x1,y1,x2,y2)],[],...]
    '''

    data_group = [[] for i in range(len(info))]
    for index in range(len(data)):
      object_boxes = data_label[index]['bbox']
      object_classes = data_label[index]['category_id']

      for ob_index in range(len(object_boxes)):
        data_group[int(object_classes[ob_index])].append((data[index],index,object_boxes[ob_index]))

    return data_group

  def filter_by_id(self, id):
    id_filter = getattr(self, 'filter', None)
    if id_filter is not None:
      if id in id_filter:
        return True
      else:
        return False

    return True

  def filter_by_condition(self, label, ext_annotation=None, ext_filter=None):
    #
    included_cls = getattr(self, 'included', None)
    excluded_cls = getattr(self, 'excluded', None)
    transform_cls = getattr(self, 'transform', None)
    # object must be larger than min_size
    min_size = getattr(self, 'min_size', None)
    # object aspect ratio must be suffer to predefined value
    # (aspect ratio = width / height)
    min_aspect_ratio = getattr(self, 'min_aspect_ratio', None)
    max_aspect_ratio = getattr(self, 'max_aspect_ratio', None)

    if included_cls is not None:
      keep_index = []
      # cls filter
      for obj_index, obj in enumerate(label['category']):
        if obj in included_cls:
          if 'bbox' in label:
            x0, y0, x1, y1 = label['bbox'][obj_index, :]
            # size filter
            if min_size is not None:
              if (x1 - x0) < min_size or (y1 - y0) < min_size:
                continue

            if min_aspect_ratio is not None:
              if float(x1 - x0) / float(y1 - y0) < min_aspect_ratio:
                continue

            if max_aspect_ratio is not None:
              if float(x1 - x0) / float(y1 - y0) > max_aspect_ratio:
                continue

          # ext filter
          if ext_filter is not None:
            is_skip = False
            for f in ext_filter:
              remained_tag = getattr(self, f, None)
              if remained_tag is not None:
                if label[f][obj_index] != remained_tag:
                  is_skip = True
                  break
            if is_skip:
              continue

          # keep obj index
          keep_index.append(obj_index)

        if len(keep_index) == 0:
          return None

        if 'bbox' in label:
          label['bbox'] = label['bbox'][keep_index, :]
        label['category_id'] = label['category_id'][keep_index]
        label['category'] = [label['category'][i] for i in keep_index]

        if ext_annotation is not None:
          for annotation_name in ext_annotation:
            if annotation_name in label:
              if type(label[annotation_name]) == list:
                label[annotation_name] = [label[annotation_name][i] for i in keep_index]
              else:
                label[annotation_name] = label[annotation_name][keep_index]

    elif excluded_cls is not None:
      keep_index = []
      for obj_index, obj in enumerate(label['category']):
        if obj not in excluded_cls:
          # size filter
          if 'bbox' in label:
            x0, y0, x1, y1 = label['bbox'][obj_index, :]
            if min_size is not None:
              if (x1 - x0) < min_size or (y1 - y0) < min_size:
                continue

            if min_aspect_ratio is not None:
              if float(x1 - x0) / float(y1 - y0) < min_aspect_ratio:
                continue

            if max_aspect_ratio is not None:
              if float(x1 - x0) / float(y1 - y0) > max_aspect_ratio:
                continue

          # ext filter
          if ext_filter is not None:
            is_skip = False
            for f in ext_filter:
              remained_tag = getattr(self, f, None)
              if remained_tag is not None:
                if label[f][obj_index] != remained_tag:
                  is_skip = True
                  break
            if is_skip:
              continue

          # keep obj index
          keep_index.append(obj_index)

      if len(keep_index) == 0:
        return None

      if 'bbox' in label:
        label['bbox'] = label['bbox'][keep_index, :]
      label['category_id'] = label['category_id'][keep_index]
      label['category'] = [label['category'][i] for i in keep_index]

      if ext_annotation is not None:
        for annotation_name in ext_annotation:
          if annotation_name in label:
            if type(label[annotation_name]) == list:
              label[annotation_name] = [label[annotation_name][i] for i in keep_index]
            else:
              label[annotation_name] = label[annotation_name][keep_index]

    if transform_cls is not None:
      for obj_index, obj in enumerate(label['category']):
        if obj in transform_cls:
          label['category_id'][obj_index] = transform_cls[obj]

    return label

  def load_samples(self, sample_folder=None):
    assert(self.train_or_test == 'sample')
    sample_file = None
    if sample_folder is None:
      sample_folder = os.path.join(self.dir, self.train_or_test)

    for file in os.listdir(sample_folder):
      if file[0] == '.':
        continue
      if file.split('.')[-1].lower() == 'sample':
        sample_file = os.path.join(sample_folder, file)
        break

    sample_data = []
    if sample_file is not None:
      with open(sample_file, 'rb') as fp:
        sample_data = loads(fp.read())
    return sample_data, list(range(len(sample_data)))

  def download(self, target_path,
               file_names=[],
               default_url=None,
               auto_untar=False,
               is_gz=False,
               auto_unzip=False,
               shell=None):
    dataset_url = getattr(self, 'dataset_url', None)
    if dataset_url is None or len(dataset_url) == 0:
      dataset_url = default_url

    if dataset_url is not None:
      # dataset dont locate local
      mltalker_root = getattr(Config, 'server_ip', None)
      is_mltalker = False

      if mltalker_root is not None and mltalker_root in dataset_url:
        terms = dataset_url.split('/')
        dataset_url = '/'.join(terms[0:-2])
        dataset_url = os.path.join(dataset_url, self.train_or_test)
        is_mltalker = True

      # validate crowdsource address
      is_http = re.match('^((https|http|ftp|rtsp|mms)?://)', dataset_url)
      if is_http is not None:
        # 3rdpart dataset
        if not os.path.exists(target_path):
          os.makedirs(target_path)

        if len(file_names) > 0:
          for file_name in file_names:
            if maybe_here(target_path, file_name) is None:
              download(os.path.join(dataset_url, file_name), target_path)

              if auto_unzip:
                with zipfile.ZipFile(os.path.join(target_path, file_name), mode='r') as zpfd:
                  zpfd.extractall(target_path)
        else:
          if not is_mltalker:
            download_path = os.path.join(target_path, dataset_url.split('/')[-1])
            if not os.path.exists(download_path):
              download(dataset_url, target_path)

              if auto_untar:
                with tarfile.open(download_path, 'r:gz' if is_gz else 'r') as tar:
                  tar.extractall(target_path)

              if auto_unzip:
                with zipfile.ZipFile(download_path, mode='r') as zpfd:
                  zpfd.extractall(target_path)
                  
              if shell is not None:
                logger.info('execute: %s' % shell)
                subprocess.call(shell, shell=True, cwd=target_path)
          else:
            download_path = os.path.join(target_path, 'data.tar')
            if not os.path.exists(download_path):
              download(dataset_url, target_path, 'data.tar')

              if auto_untar:
                with tarfile.open(download_path, 'r:gz' if is_gz else 'r') as tar:
                  tar.extractall(target_path)
              
              if shell is not None:
                logger.info('execute: %s' % shell)
                subprocess.call(shell, shell=True, cwd=target_path)
                
      # # validate ipfs address
      # is_ipfs = re.match('^(ipfs://)', dataset_url)
      # if is_ipfs:
      #   dataset_name = os.path.normpath(target_path).split('/')[-1]
      #   process = multiprocessing.Process(target=data_download_local, args=(dataset_name, dataset_url))
      #   process.start()
      #   process.join()
      #   return

      # validate shell (download and reorganize data manully)
      is_shell = re.match('^(shell:)', dataset_url)
      if is_shell is not None:
        for f in file_names:
          if os.path.exists(os.path.join(target_path, f)):
            cmd_str = dataset_url[6:]
            cmds = cmd_str.split('\n')
            for cmd in cmds:
              if '{file_placeholder}' in cmd:
                cmd = cmd.format(file_placeholder=f)
                
              logger.info('execute: %s'%cmd)
              subprocess.call(cmd, shell=True, cwd=target_path)
          else:
            logger.error('perhaps %s must be prepared manully and '
                         'put it at %s'%(f, target_path))
            os._exit(-1)


class UnlabeledDataset(Dataset):
  def __init__(self, dataset):
    super(UnlabeledDataset, self).__init__()
    self.dataset = dataset
    self.unlabeled_data = dataset.unlabeled()
    self.dir = dataset.dir

  def data_pool(self):
    for id in range(self.size):
      yield self.at(id)

  def at(self, unlabeled_id):
    id = self.unlabeled_data[unlabeled_id]['id']
    file_id = self.unlabeled_data[unlabeled_id]['file_id']
    data = self.dataset.at(id, file_id)
    data[1].update({'file_id': file_id, 'id': id})
    return data

  @property
  def size(self):
    return len(self.unlabeled_data)


class CandidateDataset(Dataset):
  def __init__(self, dataset):
    super(CandidateDataset, self).__init__()
    self.dataset = dataset
    self.dir = dataset.dir
    self.candidates = []

    # 使用candidate文件夹中的数据(已经完成标注的数据)
    if os.path.exists(self.dataset.candidate_file):
      with open(self.dataset.candidate_file, 'r') as fp:
        content = fp.readline()
        content = content.strip()

        while content != '':
          id, file_id, annotation_file_id = content.split(',')

          self.candidates.append({
            'id': id,
            'file_id': file_id,
            'annotation_file_id': annotation_file_id
          })
          content = fp.readline()
          content = content.strip()

  def data_pool(self):
    for id in range(self.size):
      yield self.at(id)

  def at(self, local_id):
    id = self.candidates[local_id]['id']
    file_id = self.candidates[local_id]['file_id']
    annotation_file_id = self.candidates[local_id]['annotation_file_id']
 
    data = self.dataset.at((int)(id), file_id)
    data[1].update({'file_id': file_id, 'id': id, 'annotation_file_id': os.path.join(self.dataset.dir, annotation_file_id)})

    return data

  def split(self, split_params, split_method):
    _, b = self.dataset.split(split_params, split_method)
    return self, b

  @property
  def size(self):
    return len(self.candidates)


class DataAnnotationSplitDataset(object):
  def __init__(self, dataset):
      super().__init__()
      self.dataset = dataset
    
  def data_pool(self):
    pass

  def at(self, id):
    self.dataset.at(id)
    pass

  def annotation_at(self, id):
    
    return

  @property
  def size(self):
    return len(self.dataset)

  def waiting_process_num(self):
    return self.dataset.waiting_process_num()

  def finish_process_num(self):
    return self.dataset.finish_process_num()



import torch
from antgo.framework.helper.models.builder import MODELS, build_model
from antgo.framework.helper.multi_stream_module import MultiSteamModule
from antgo.framework.helper.models.ema_module import *


@MODELS.register_module()
class MPL(MultiSteamModule):
    def __init__(self, model: dict, ema=0.0, train_cfg=None, test_cfg=None, init_cfg=None) -> None:
        if test_cfg is None:
            test_cfg = dict()
        test_cfg.update(
            {
                'inference_on': 'student'
            }
        )
        super(MPL, self).__init__(
            dict(
                teacher=build_model(model['teacher']),
                student=build_model(model['student']),
            ),
            train_cfg=train_cfg if train_cfg is not None else dict(),
            test_cfg=test_cfg if test_cfg is not None else dict(),
        )

        self.avg_student = None
        if ema > 0:
            self.avg_student = ModelEMA(self.student, ema)
            self.inference_on = self.avg_student

        self.temperature = train_cfg.get('temperature', 0.7)
        self.lambda_u = train_cfg.get('lambda_u', 8.0)
        self.uda_steps = train_cfg.get('uda_steps', 5000)
        self.grad_clip = train_cfg.get('grad_clip', 1e9)
        self.ema = train_cfg.get('ema', 0)
        self.threshold = train_cfg.get('threshold', 0.6)
        self.label_batch_size = train_cfg.get('label_batch_size', 5)
        self.label_smoothing = train_cfg.get('label_smoothing', 0.15)
        self.criterion = nn.CrossEntropyLoss(label_smoothing=self.label_smoothing)

    def uda_loss(self, logits_uw, logits_us):
        # logits_uw: 弱增强获得的结果
        # logits_us: 强增强获得的结果
        mask, soft_pseudo_label, hard_pseudo_label = self.make_pseudo_label(logits_uw)

        loss_u = 0
        if len(logits_us.shape) == 2:
            loss_u = torch.mean(
                    -(soft_pseudo_label * torch.log_softmax(logits_us, dim=-1)).sum(dim=-1) * mask
                )
        else:
            print('incoming')
            pass
        return mask, hard_pseudo_label, loss_u

    def train_step(self, data, optimizer, **kwargs):
        image = data.pop('image')
        other = data
        label_batch_size = self.label_batch_size
        batch_size = image.shape[0]
        unlabel_batch_size = (batch_size - label_batch_size)//2

        label_images, unlabel_ws_images = \
            torch.split(image, [label_batch_size, batch_size-label_batch_size],dim=0) 

        unlabel_w_images=torch.index_select(unlabel_ws_images,dim=0,index=torch.tensor([i for i in range(0,2*unlabel_batch_size,2)]))
        unlabel_s_images=torch.index_select(unlabel_ws_images,dim=0,index=torch.tensor([i for i in range(1,2*unlabel_batch_size,2)]))

        label_kwargs = {}
        unlabel_w_kwargs = {}
        unlabel_s_kwargs = {}
        for k, v in other.items():
            if isinstance(v, list):
                v_label_kwargs = v[:label_batch_size]
                v_unlabel_ws_kwargs = v[v_label_kwargs:]
                v_unlabel_w_kwargs = [v_unlabel_ws_kwargs[i] for i in range(0, 2*unlabel_batch_size, 2)]
                v_unlabel_s_kwargs = [v_unlabel_ws_kwargs[i] for i in range(1, 2*unlabel_batch_size, 2)]
                
                unlabel_w_kwargs[k] = v_unlabel_w_kwargs
                unlabel_s_kwargs[k] = v_unlabel_s_kwargs
                continue

            v_label_kwargs, v_unlabel_ws_kwargs = \
                torch.split(v, [label_batch_size, unlabel_batch_size * 2],dim=0) 
            label_kwargs[k] = v_label_kwargs

            v_unlabel_w_kwargs=torch.index_select(v_unlabel_ws_kwargs,dim=0,index=torch.tensor([i for i in range(0,2*unlabel_batch_size,2)]))
            v_unlabel_s_kwargs=torch.index_select(v_unlabel_ws_kwargs,dim=0,index=torch.tensor([i for i in range(1,2*unlabel_batch_size,2)]))

            unlabel_w_kwargs[k] = v_unlabel_w_kwargs
            unlabel_s_kwargs[k] = v_unlabel_s_kwargs

        t_images = torch.cat((label_images, unlabel_w_images, unlabel_s_images))
        t_logits = self.teacher.forward_train(t_images)
        t_logits_l = t_logits[:label_batch_size]
        t_logits_uw, t_logits_us = t_logits[label_batch_size:].chunk(2)
        del t_logits
        
        t_loss_dict = self.teacher.loss(t_logits_l, **label_kwargs)
        t_loss_l = sum(_value for _, _value in t_loss_dict.items())

        # 具体如何基于weak sample和strong sample构建损失，交给MPL自定义
        # soft_pseudo_label = torch.softmax(t_logits_uw.detach() / self.model.temperature, dim=-1)
        # max_probs, hard_pseudo_label = torch.max(soft_pseudo_label, dim=-1)
        # mask = max_probs.ge(self.model.threshold).float()
        # t_loss_u = torch.mean(
        #         -(soft_pseudo_label * torch.log_softmax(t_logits_us, dim=-1)).sum(dim=-1) * mask
        #     )
        mask, hard_pseudo_label, t_loss_u = self.uda_loss(t_logits_uw.detach() / self.temperature, t_logits_us)

        weight_u = self.lambda_u * min(1., (kwargs['iter'] + 1) / self.uda_steps)
        t_loss_uda = t_loss_l + weight_u * t_loss_u

        s_images = torch.cat((label_images, unlabel_s_images))
        s_logits = self.student.forward_train(s_images)
        s_logits_l = s_logits[:label_batch_size]
        s_logits_us = s_logits[label_batch_size:]
        del s_logits

        s_loss_dict_old = self.student.loss(s_logits_l.detach(), **label_kwargs)
        s_loss_l_old = sum(_value for _, _value in s_loss_dict_old.items())
        s_loss = self.criterion(s_logits_us, hard_pseudo_label)
        
        s_loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.student.parameters(), self.grad_clip)
        
        optimizer['student'].step()
        if self.ema > 0:
            self.avg_student.update_parameters(self.student)     

        ####
        with torch.no_grad():
            s_logits_l = self.student.forward_train(label_images)
            s_loss_dict_new = self.student.loss(s_logits_l.detach(), **label_kwargs)
            s_loss_l_new = sum(_value for _, _value in s_loss_dict_new.items())

        # theoretically correct formula (https://github.com/kekmodel/MPL-pytorch/issues/6)
        # dot_product = s_loss_l_old - s_loss_l_new

        # author's code formula
        dot_product = s_loss_l_new - s_loss_l_old
        # moving_dot_product = moving_dot_product * 0.99 + dot_product * 0.01
        # dot_product = dot_product - moving_dot_product

        # TODO，这里感觉应该改成使用弱增强来获得伪标签
        _, _, hard_pseudo_label = self.make_pseudo_label(t_logits_us.detach())
        t_loss_mpl = dot_product * self.criterion(t_logits_us, hard_pseudo_label)
        # test
        # t_loss_mpl = torch.tensor(0.).to(args.device)
        t_loss = t_loss_uda + t_loss_mpl
        t_loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.teacher.parameters(), self.grad_clip)
        
        optimizer['teacher'].step()
        self.teacher.zero_grad()
        self.student.zero_grad()

        losses = {
            's_loss': s_loss,
            't_loss': t_loss,
            't_loss_l': t_loss_l,
            't_loss_u': t_loss_u,
            't_loss_mpl': t_loss_mpl,            
        }

        losses, log_vars = self._parse_losses(losses)
        outputs = dict(
            loss=losses, log_vars=log_vars, num_samples=batch_size)

        outputs['log_vars'].update(
            {'mask': mask.mean().cpu().item()}
        )
        return outputs

    def make_pseudo_label(self, logits_u):
        # 返回 mask, soft_pseudo_label, hard_pseudo_label
        # NxC
        soft_pseudo_label = torch.softmax(logits_u, dim=-1)
        max_probs, hard_pseudo_label = torch.max(soft_pseudo_label, dim=-1)
        mask = max_probs.ge(self.threshold).float()
        return mask, soft_pseudo_label, hard_pseudo_label

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs):
        # 同时兼容从base模型和dense模型加载
        if not any(["student" in key or "teacher" in key for key in state_dict.keys()]):
            keys = list(state_dict.keys())
            state_dict.update({"teacher." + k: state_dict[k] for k in keys})
            state_dict.update({"student." + k: state_dict[k] for k in keys})
            for k in keys:
                state_dict.pop(k)

        return super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )


from .extract import (extract_from_videos, extract_from_images, extract_from_samples, extract_from_crop, extract_from_coco)
from .browser_funcs import (browser_images)
from .filter_funcs import (filter_by_tags)
from .package import (package_to_kv, package_to_tfrecord)
from .download_funcs import (download_from_baidu, download_from_bing, download_from_vcg)
from .label_funcs import (label_to_studio, label_from_studio, label_start, label_from_native, label_to_merge)
from .share_funcs import (share_data_func)
__all__ = [
    'extract_from_videos', 
    'extract_from_images', 
    'extract_from_samples', 
    'extract_from_crop',
    'extract_from_coco',
    'browser_images', 
    'filter_by_tags',
    'package_to_kv',
    'package_to_tfrecord',
    'download_from_baidu',
    'download_from_bing',
    'download_from_vcg',
    'label_to_studio',
    'label_from_studio',
    'label_from_native',
    'label_start',
    'label_to_merge',
    'share_data_func'
]